# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil

from pisek.env.env import Env
from pisek.config.config_types import ProgramRole
from pisek.config.task_config import RunSection
from pisek.utils.paths import TaskPath, LogPath
from pisek.task_jobs.program import RunResultKind
from pisek.task_jobs.data.testcase_info import TestcaseInfo

from .base_classes import GeneratorListInputs, GenerateInput


class CmsOldListInputs(GeneratorListInputs):
    """Lists all inputs for cms-old generator - by running it."""

    def __init__(self, env: Env, generator: RunSection, **kwargs) -> None:
        super().__init__(env, generator, name="Run generator", **kwargs)

    def _run(self) -> list[TestcaseInfo]:
        """Generates all inputs."""
        gen_dir = TaskPath.generated_path(".")
        try:
            shutil.rmtree(gen_dir.path)
        except FileNotFoundError:
            pass
        self.makedirs(gen_dir, exist_ok=False)

        run_result = self._run_program(
            ProgramRole.gen,
            self.generator,
            args=[gen_dir.abspath],
            stderr=LogPath.generator_log(self.generator.name),
        )
        self._access_dir(gen_dir)

        if run_result.kind != RunResultKind.OK:
            raise self._create_program_failure(
                "Generator failed:",
                run_result,
                stderr_force_content=True,
            )

        testcases = []
        for inp in self._globs_to_files(["*"], TaskPath.generated_path(".")):
            if inp.path.endswith(".in"):
                testcases.append(
                    TestcaseInfo.generated(inp.name.removesuffix(".in"), seeded=False)
                )
        return testcases


class CmsOldGenerate(GenerateInput):
    def __init__(
        self,
        env: Env,
        generator: RunSection,
        testcase_info: TestcaseInfo,
        seed: int,
        **kwargs,
    ) -> None:
        super().__init__(
            env,
            generator,
            testcase_info,
            seed,
            name=f"Serve {testcase_info.input_path().name}",
            **kwargs,
        )

    def _gen(self):
        self._link_file(
            TaskPath.generated_path(self.input_path.name),
            self.input_path.to_raw(self._env.config.tests.in_format),
            overwrite=True,
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import abstractmethod
from decimal import Decimal
import subprocess
from typing import Optional

from pisek.env.env import Env
from pisek.utils.paths import TaskPath, IInputPath, IOutputPath
from pisek.jobs.jobs import PipelineItemFailure
from pisek.utils.text import tab
from pisek.task_jobs.run_result import RunResult, RunResultKind
from pisek.task_jobs.solution.solution_result import (
    Verdict,
    SolutionResult,
    RelativeSolutionResult,
)
from pisek.task_jobs.checker.checker_base import RunBatchChecker


class RunJudgeLibChecker(RunBatchChecker):
    """Checks solution output and correct output using judgelib checker."""

    @abstractmethod
    def _get_flags(self) -> list[str]:
        pass

    def _check(self) -> SolutionResult:
        self._access_file(self.output)
        self._access_file(self.correct_output)

        checker = self._run_tool(
            "_" + self.checker_name,
            args=[
                *self._get_flags(),
                self.output.abspath,
                self.correct_output.abspath,
            ],
            stderr=self.checker_log_file,
        )

        stderr: str = self._read_file(self.checker_log_file)

        # XXX: Okay, it didn't finish in no time, but this is not meant to be used
        rr = RunResult(
            RunResultKind.OK,
            checker.returncode,
            0,
            0,
            status=(stderr.strip() or "Files are equivalent")
            + f": {self.output.col(self._env)} {self.correct_output.col(self._env)}",
        )

        if checker.returncode == 42:
            return RelativeSolutionResult(
                verdict=Verdict.ok,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=rr,
                relative_points=Decimal(1),
            )
        elif checker.returncode == 43:
            return RelativeSolutionResult(
                verdict=Verdict.wrong_answer,
                message=None,
                solution_rr=self._solution_run_res,
                checker_rr=rr,
                relative_points=Decimal(0),
            )
        else:
            raise PipelineItemFailure(f"{self.checker_name} failed:\n{tab(stderr)}")


class RunTokenChecker(RunJudgeLibChecker):
    """Checks solution output and correct output using judge-token."""

    def __init__(
        self,
        env: Env,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
    ) -> None:
        super().__init__(
            env=env,
            checker_name="judge-token",
            test=test,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_verdict=expected_verdict,
        )

    def _get_flags(self) -> list[str]:
        flags = ["-t"]
        if self._env.config.tests.tokens_ignore_newlines:
            flags.append("-n")
        if self._env.config.tests.tokens_ignore_case:
            flags.append("-i")
        if self._env.config.tests.tokens_float_rel_error != None:
            flags.extend(
                [
                    "-r",
                    "-e",
                    str(self._env.config.tests.tokens_float_rel_error),
                    "-E",
                    str(self._env.config.tests.tokens_float_abs_error),
                ]
            )
        return flags


class RunShuffleChecker(RunJudgeLibChecker):
    """Checks solution output and correct output using judge-shuffle."""

    def __init__(
        self,
        env: Env,
        test: int,
        input_: IInputPath,
        output: IOutputPath,
        correct_output: IOutputPath,
        expected_verdict: Optional[Verdict],
    ) -> None:
        super().__init__(
            env=env,
            checker_name="judge-shuffle",
            test=test,
            input_=input_,
            output=output,
            correct_output=correct_output,
            expected_verdict=expected_verdict,
        )

    def _get_flags(self) -> list[str]:
        SHUFFLE_MODE_FLAGS = {
            "lines": "-l",
            "words": "-w",
            "lines_words": "-lw",
            "tokens": "-nw",
        }
        assert self._env.config.tests.shuffle_mode is not None
        flags = ["-e", SHUFFLE_MODE_FLAGS[self._env.config.tests.shuffle_mode]]
        if self._env.config.tests.shuffle_ignore_case:
            flags.append("-i")

        return flags



import re
from colorama import Fore, Back

_ANSI_REGEX = re.compile(r"\x1b\[[0-9;]*[A-Za-z]")


def remove_colors(text: str) -> str:
    return _ANSI_REGEX.sub("", text)


class __ColorSettings:
    """Singleton object to store current color settings."""

    def __init__(self) -> None:
        self.colors_on = True

    def set_state(self, colors_on: bool) -> None:
        """Sets whether colors should be displayed."""
        self.colors_on = colors_on

    def _colored(self, what, msg: str, color: str) -> str:
        """Recolors all white text to given color."""
        if not self.colors_on:
            return msg

        col = getattr(what, color.upper())
        msg = msg.replace(f"{what.RESET}", f"{what.RESET}{col}")
        return f"{col}{msg}{what.RESET}"

    def colored(self, msg: str, color: str) -> str:
        return self._colored(Fore, msg, color)

    def colored_back(self, msg: str, color: str) -> str:
        return self._colored(Back, msg, color)


ColorSettings = __ColorSettings()



from abc import ABC, abstractmethod
from difflib import SequenceMatcher
from functools import partial
from importlib.resources import files
import re
from typing import TYPE_CHECKING, Callable, Iterable, Optional

from pisek.utils.text import tab
from pisek.user_errors import TaskConfigError

if TYPE_CHECKING:
    from .config_hierarchy import ConfigHierarchy

CONFIG_DESCRIPTION = str(files("pisek").joinpath("config/config-description"))


def basic_similarity(a: str, b: str) -> float:
    return SequenceMatcher(a=a, b=b).ratio()


def regex_score(regex: str, name: str) -> float:
    if re.match(regex, name):
        return 1.0
    else:
        # XXX: Very incomplete and fragile, but it works
        regex_str = regex.replace(r"(.*)", "\x00").replace(r"\d{2}", "\x00")
        return basic_similarity(regex_str, name)


class ApplicabilityCondition(ABC):
    @abstractmethod
    def check(self, section: str, key: str, config: "ConfigHierarchy") -> str:
        pass


class KeyValueCond(ApplicabilityCondition):
    def __init__(
        self,
        section: "ConfigSectionDescription",
        key: "ConfigKeyDescription",
        value: str,
    ) -> None:
        self.section = section
        self.key = key
        self.value = value
        super().__init__()

    @abstractmethod
    def _op(self, current_value) -> bool:
        pass

    def check(self, section: str, key: str, config: "ConfigHierarchy") -> str:
        section = self.section.transform_name(section)
        key = self.key.transform_name(key)
        current_value = self.key.get(section, key, config)
        if self._op(current_value):
            return ""
        return f"[{section}] {key}={current_value}\n"


class KeyHasValue(KeyValueCond):
    def _op(self, current_value):
        return self.value == current_value


class KeyHasNotValue(KeyValueCond):
    def _op(self, current_value):
        return self.value != current_value


class ConfigSectionDescription:
    def __init__(self, section: str) -> None:
        self.section = section
        self.defaults_to: list[str] = []
        self.dynamic_default: bool = False
        self.similarity_function: Optional[Callable[[str], float]] = None

    def similarity(self, section: str) -> float:
        if self.similarity_function is None:
            return basic_similarity(self.section, section)
        else:
            return self.similarity_function(section)

    def transform_name(self, name: str) -> str:
        return name if self.similarity(name) == 1.0 else self.section


class ConfigKeyDescription:
    def __init__(self, section: ConfigSectionDescription, key: str) -> None:
        self.section = section
        self.key = key
        self.defaults_to: list[tuple[str, str]] = []
        self.dynamic_default: bool = False
        self.applicability_conditions: list[ApplicabilityCondition] = []
        self.similarity_function: Optional[Callable[[str], float]] = None

    def get(self, section: str, key: str, config: "ConfigHierarchy") -> str:
        return config.get_from_candidates(
            [(self.section.transform_name(section), self.transform_name(key))]
            + self.defaults()
        ).value

    def defaults(self) -> list[tuple[str, str]]:
        if self.dynamic_default:
            raise NotImplementedError("Dynamic defaulting not implemented")
        return self.defaults_to + [(d, self.key) for d in self.section.defaults_to]

    def similarity(self, key: str) -> float:
        if self.similarity_function is None:
            return basic_similarity(self.key, key)
        else:
            return self.similarity_function(key)

    def score(self, section: str, key: str) -> float:
        return (9 * self.section.similarity(section) + self.similarity(key)) / 10

    def applicable(self, section: str, key: str, config: "ConfigHierarchy") -> str:
        text = ""
        for cond in self.applicability_conditions:
            text += cond.check(section, key, config)
        return text

    def transform_name(self, name: str) -> str:
        return name if self.similarity(name) == 1.0 else self.key


class ConfigKeysHelper:
    def __init__(self) -> None:
        self.sections: dict[str, ConfigSectionDescription] = {}
        self.keys: dict[tuple[str, str], ConfigKeyDescription] = {}
        self.key_index: dict[str, list[ConfigKeyDescription]] = {}
        add_applicability_conditions: list[
            tuple[ConfigKeyDescription, Callable[[], ApplicabilityCondition]]
        ] = []
        with open(CONFIG_DESCRIPTION) as f:
            section: Optional[ConfigSectionDescription] = None
            last_key: Optional[ConfigKeyDescription] = None
            for line in f:
                line = line.strip()

                if len(line) == 0:
                    pass

                elif line.startswith("#!"):
                    assert section is not None
                    [fun, *args] = line.removeprefix("#!").split()
                    if fun == "regex":
                        if last_key is None:
                            section.similarity_function = partial(regex_score, args[0])
                        else:
                            last_key.similarity_function = partial(regex_score, args[0])
                    elif fun == "if":
                        assert section is not None
                        assert last_key is not None

                        if len(args) == 2 and args[1].count("==") == 1:
                            key_name, value = args[1].split("==")
                            add_applicability_conditions.append(
                                (
                                    last_key,
                                    self._gen_key_has_value(args[0], key_name, value),
                                )
                            )
                        elif len(args) == 2 and args[1].count("!=") == 1:
                            key_name, value = args[1].split("!=")
                            add_applicability_conditions.append(
                                (
                                    last_key,
                                    self._gen_key_has_not_value(
                                        args[0], key_name, value
                                    ),
                                )
                            )
                        else:
                            self._invalid_function_args(fun, args)
                    elif fun == "default":
                        if last_key is None:
                            if len(args) != 1:
                                self._invalid_function_args(fun, args)
                            section.defaults_to.append(args[0])
                        else:
                            if len(args) != 2:
                                self._invalid_function_args(fun, args)
                            last_key.defaults_to.append((args[0], args[1]))
                    elif fun == "dynamic_default":
                        if last_key is None:
                            if len(args) != 0:
                                self._invalid_function_args(fun, args)
                            section.dynamic_default = True
                        else:
                            if len(args) != 0:
                                self._invalid_function_args(fun, args)
                            last_key.dynamic_default = True
                    else:
                        raise ValueError(
                            f"invalid config-description function: '{fun}'"
                        )

                elif line.count("=") == 1 and line[-1] == "=":
                    assert section is not None
                    last_key = ConfigKeyDescription(section, line.removesuffix("="))
                    self.keys[(section.section, last_key.key)] = last_key
                    self.key_index.setdefault(last_key.key, []).append(last_key)

                elif line[0] == "[" and line[-1] == "]":
                    last_key = None
                    section = ConfigSectionDescription(
                        line.removeprefix("[").removesuffix("]")
                    )
                    self.sections[section.section] = section

                elif line.startswith("#"):
                    pass
                else:
                    raise ValueError(f"invalid config-description line: '{line}'")

        for key, lambda_cond in add_applicability_conditions:
            key.applicability_conditions.append(lambda_cond())

    def _gen_key_has_value(
        self, section: str, key: str, value: str
    ) -> Callable[[], KeyValueCond]:
        return lambda: KeyHasValue(
            self.sections[section], self.keys[(section, key)], value
        )

    def _gen_key_has_not_value(
        self, section: str, key: str, value: str
    ) -> Callable[[], KeyValueCond]:
        return lambda: KeyHasNotValue(
            self.sections[section], self.keys[(section, key)], value
        )

    def _invalid_function_args(self, fun_name: str, args: list[str]) -> None:
        raise ValueError(
            f"invalid config-description function {fun_name} arguments: '{' '.join(args)}'"
        )

    def find_section(self, section: str) -> tuple[float, str]:
        return max((s.similarity(section), s.section) for s in self.sections.values())

    def find_key(
        self,
        section: str,
        key: str,
        config: "ConfigHierarchy",
        allow_unapplicable: bool,
    ) -> tuple[float, str, str]:
        keys: Iterable[ConfigKeyDescription]
        if (
            key in self.key_index
            and max(k.score(section, key) for k in self.key_index[key]) == 1.0
        ):
            keys = self.key_index[key]
        else:
            keys = self.keys.values()

        for candidate in keys:
            if candidate.score(section, key) == 1.0:
                if not allow_unapplicable and (
                    text := candidate.applicable(section, key, config)
                ):
                    raise TaskConfigError(
                        f"Key '{key}' not allowed in this context:\n{tab(text).rstrip()}"
                    )
                else:
                    return (1.0, section, candidate.key)

        recommendation = max(
            (
                k.score(section, key),
                k.section.transform_name(section),
                k.transform_name(key),
            )
            for k in self.keys.values()
        )
        assert 0.0 <= recommendation[0] < 1.0
        return recommendation



"""
Tests the command-line interface.
"""

import os

import unittest
from io import StringIO
from unittest import mock

from util import TestFixture

from pisek.__main__ import main


class TestCLI(TestFixture):
    @property
    def fixture_path(self) -> str:
        return "fixtures/sum_cms/"

    def args(self) -> list[list[str]]:
        return [["test", "--time-limit", "0.2"]]

    def runTest(self) -> None:
        if not self.fixture_path:
            return

        self.log_files()

        with mock.patch("sys.stdout", new=StringIO()) as std_out:
            with mock.patch("sys.stderr", new=StringIO()) as std_err:
                for args_i in self.args():
                    result = main(args_i)

                    self.assertFalse(
                        result,
                        f"Command failed: {' '.join(args_i)}",
                    )

        self.check_files()


class TestCLITestSolution(TestCLI):
    def args(self) -> list[list[str]]:
        return [["test", "solutions", "solve"]]


class TestCLITestGenerator(TestCLI):
    def args(self) -> list[list[str]]:
        return [["test", "generator"]]


class TestCLIClean(TestCLI):
    def args(self) -> list[list[str]]:
        return [["clean"]]


class TestCLITestingLog(TestCLI):
    def args(self) -> list[list[str]]:
        return [["test", "--testing-log"]]

    def created_files(self) -> list[str]:
        return ["testing_log.json"]


class TestCLIVisualize(TestCLI):
    def args(self) -> list[list[str]]:
        return [["test", "--testing-log"], ["visualize"]]

    def created_files(self) -> list[str]:
        return ["testing_log.json"]


class TestCLIExport(TestCLI):
    @property
    def fixture_path(self) -> str:
        return "fixtures/guess/"

    def args(self) -> list[list[str]]:
        return [["config", "export", "config"], ["test"]]

    def created_files(self) -> list[str]:
        return ["exported-config"]


if __name__ == "__main__":
    unittest.main(verbosity=2)



from configparser import ConfigParser
from functools import partial
from importlib.resources import files
import os
import shutil
from typing import Callable, NoReturn, Sequence

from pisek.user_errors import InvalidArgument, InvalidOperation
from pisek.utils.colors import ColorSettings
from pisek.utils.user_input import input_string, input_choice
from pisek.config.config_types import (
    TaskType,
    GenType,
    ValidatorType,
    OutCheck,
    JudgeType,
)
from pisek.config.config_hierarchy import DEFAULT_CONFIG_FILENAME
from pisek.task_jobs.program import ProgramRole

EXAMPLE_TASKS_DIR = str(files("pisek").joinpath("../examples"))
EXAMPLE_TASKS = ["cms-batch", "opendata"]

OFFERED_GEN_TYPES = [GenType.pisek_v1, GenType.opendata_v1, GenType.cms_old]
RECOMMENDED_GEN_TYPES = [GenType.pisek_v1]

OFFERED_VALIDATOR_TYPES = [ValidatorType.simple_42]
RECOMMENDED_VALIDATOR_TYPES: list[ValidatorType] = (
    []
)  # All offered validator types are reasonable, no need to recommend anything

OFFERED_JUDGE_TYPES = {
    TaskType.batch: [JudgeType.cms_batch, JudgeType.opendata_v2],
    TaskType.interactive: [JudgeType.cms_communication],
}
RECOMMENDED_JUDGE_TYPES: dict[TaskType, list[JudgeType]] = (
    {  # All offered judge types are reasonable, no need to recommend anything
        TaskType.batch: [],
        TaskType.interactive: [],
    }
)

SOLUTION_SUBDIR = "solutions"


def remove_suffix(path: str) -> str:
    return os.path.splitext(path)[0]


def touch(path: str) -> None:
    open(path, "a").close()


def recommended() -> str:
    return " (" + ColorSettings.colored("recommended", "green") + ")"


def invalid_config_name(config_filename: str) -> NoReturn:
    raise InvalidArgument(f"Config filename '{config_filename}' already in use.")


def input_program(
    p_role: ProgramRole,
    types: Sequence[str],
    recommended_types: Sequence[str],
    no_jumps: bool,
) -> tuple[str, str]:

    choices = [
        (t, t + (recommended() if t in recommended_types else "")) for t in types
    ]
    type_ = input_choice(f"Choose {p_role}_type:", choices, no_jumps)
    filename = input_string(f"Enter {p_role} filename: ")
    print()
    return type_, filename


def init_task(config_filename: str, no_jumps: bool) -> None:
    if os.listdir():
        raise InvalidOperation("Current directory is not empty")

    example_tasks: list[tuple[Callable[[str], None], str]] = [
        (
            partial(from_template, os.path.join(EXAMPLE_TASKS_DIR, task)),
            task + " example task",
        )
        for task in EXAMPLE_TASKS
    ]
    _from_scratch: Callable[[str], None] = partial(from_scratch, no_jumps=no_jumps)
    create = input_choice(
        "Create a task", [(_from_scratch, "From scratch")] + example_tasks, no_jumps
    )
    create(config_filename)


def from_scratch(config_filename: str, no_jumps: bool) -> None:
    config = ConfigParser(interpolation=None)
    config.add_section("task")
    config.add_section("tests")
    config["task"]["version"] = "v3"

    # --- task_type ---
    print()
    task_type = input_choice("Choose task_type:", [(t, t) for t in TaskType], no_jumps)
    config["task"]["task_type"] = task_type
    print()

    # --- inputs ---
    gen_type, gen_name = input_program(
        ProgramRole.gen, OFFERED_GEN_TYPES, RECOMMENDED_GEN_TYPES, no_jumps
    )
    touch(gen_name)
    config["tests"]["in_gen"] = remove_suffix(gen_name)
    config["tests"]["gen_type"] = gen_type

    touch("sample.in")
    touch("sample.out")

    # --- validator ---
    val_type, val_name = input_program(
        ProgramRole.validator,
        OFFERED_VALIDATOR_TYPES,
        RECOMMENDED_VALIDATOR_TYPES,
        no_jumps,
    )
    touch(val_name)
    config["tests"]["validator"] = remove_suffix(val_name)
    config["tests"]["validator_type"] = val_type

    # --- out_check ---
    if task_type == TaskType.batch:
        out_check = input_choice(
            "How to check outputs?",
            [
                (OutCheck.tokens, "tokens - output is unique"),
                (OutCheck.shuffle, "shuffle - output is unique up to order"),
                (OutCheck.judge, "judge - output is not unique"),
            ],
            no_jumps,
        )
        print()
    else:
        out_check = OutCheck.judge

    config["tests"]["out_check"] = out_check

    if out_check == OutCheck.judge:
        judge_type, judge_name = input_program(
            ProgramRole.judge,
            OFFERED_JUDGE_TYPES[task_type],
            RECOMMENDED_JUDGE_TYPES[task_type],
            no_jumps,
        )
        touch(judge_name)
        config["tests"]["out_judge"] = remove_suffix(judge_name)
        config["tests"]["judge_type"] = judge_type

    # --- solutions ---
    os.makedirs(SOLUTION_SUBDIR)
    config.add_section("run_solution")
    config["run_solution"]["time_limit"] = "1"
    config["run_solution"]["subdir"] = "solutions"

    sol = input_string("Enter primary solution filename: ")
    touch(os.path.join(SOLUTION_SUBDIR, sol))
    sol_sec = f"solution_{remove_suffix(sol)}"
    config.add_section(sol_sec)
    config[sol_sec]["primary"] = "yes"

    try:
        with open(config_filename, "x") as f:
            config.write(f, space_around_delimiters=False)
    except FileExistsError:
        invalid_config_name(config_filename)

    print()
    print("For more information visit our docs: https://piskoviste.github.io/pisek/")


def from_template(path: str, config_filename: str) -> None:
    if config_filename != DEFAULT_CONFIG_FILENAME and os.path.exists(
        os.path.join(path, config_filename)
    ):
        invalid_config_name(config_filename)

    for item in os.listdir(path):
        s = os.path.join(path, item)
        if os.path.isdir(s):
            shutil.copytree(s, item)
        else:
            shutil.copy(s, item)

    shutil.move(DEFAULT_CONFIG_FILENAME, config_filename)



#!/usr/bin/env python3
import itertools
import random
import sys


def gen(test: int) -> None:
    max_n = 10**5
    max_k = 10**5 if test == 2 else 10**9

    # Dividing max_n by four actually makes stronger tests
    # as sandcastles of size 1 doesn't use half of the sand
    n = random.randint(1, max_n // 4)
    k = random.randint(n, max_k)

    if test == 1:
        k = n

    print(n, k)


random.seed(sys.argv[2])
gen(int(sys.argv[1]))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Antonín Maloň <git@tonyl.eu>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from dataclasses import dataclass
from decimal import Decimal
import json
from math import ceil, inf
import os
from typing import Any, Optional

from pisek.user_errors import MissingFile
from pisek.utils.text import pad, tab, eprint
from pisek.utils.colors import ColorSettings
from pisek.utils.terminal import terminal_width
from pisek.config.config_types import TestPoints
from pisek.config.task_config import load_config, TaskConfig
from pisek.config.select_solutions import expand_solutions
from pisek.task_jobs.solution.solution_result import Verdict
from pisek.task_jobs.solution.verdicts_eval import check_verdicts


class MissingSolution(Exception):
    def __init__(self, name: str) -> None:
        super().__init__(f"Missing solution {name} in testing log.")


@dataclass
class LoggedResult:
    verdict: Verdict
    relative_points: Optional[Decimal]
    absolute_points: Optional[Decimal]
    time: float
    test: str
    original_verdict: Verdict

    def points(self, test_points: TestPoints) -> Decimal:
        if test_points == "unscored":
            return Decimal("0")
        if self.relative_points is not None:
            return self.relative_points * test_points
        elif self.absolute_points is not None:
            return self.absolute_points
        else:
            raise ValueError("Relative and absolute points unset")

    def to_str(
        self,
        limit: float,
        segments: int,
        test_pad_length: int = 15,
    ) -> str:
        def get_bar() -> str:
            percentage = self.time / limit
            full = max(ceil(segments * percentage), 1)

            bounded = min(full, segments)
            overflown = max(0, full - segments)

            overflown_max_length = terminal_width - segments - test_pad_length - 40
            if cut := (overflown > overflown_max_length):
                overflown = overflown_max_length

            if abs(percentage - 1) < 0.1:
                color = "red"
            elif abs(percentage - 1) < 0.5:
                color = "yellow"
            else:
                color = "green"

            full_bar = ColorSettings.colored("━", color)

            return (
                f"[{full_bar*bounded}{('━' if ColorSettings.colors_on else ' ')*(segments-bounded)}|"
                f"{full_bar*overflown}{' '*(overflown_max_length - overflown)}{'⋯⋯' if cut else '  '}"
            )

        return (
            f"{pad(self.test, test_pad_length)}  "
            f"{self.original_verdict.mark}->{self.verdict.mark}  "
            f"{get_bar()}    {self.time:.2f} / {limit:.2f}s"
        )


def limit_result(result: LoggedResult, limit: float) -> LoggedResult:
    if result.time <= limit and result.verdict == Verdict.timeout:
        new_verdict = result.original_verdict
        if new_verdict == Verdict.timeout:
            new_verdict = Verdict.ok

        return LoggedResult(
            new_verdict,
            Decimal(1),
            None,
            result.time,
            result.test,
            result.original_verdict,
        )
    if result.time > limit:
        return LoggedResult(
            Verdict.timeout,
            Decimal(0),
            None,
            result.time,
            result.test,
            result.original_verdict,
        )
    return result


class SolutionResults:
    def __init__(
        self, name: str, config: TaskConfig, results: list[LoggedResult]
    ) -> None:
        self.name = name
        self._solution = config.solutions[name]
        self._config = config
        self._results = results

        self._results.sort(
            key=lambda r: (self._get_test(r), r.verdict.value, r.time, r.test)
        )

    def _get_test(self, result: LoggedResult) -> int:
        return min(
            i
            for i, sub in self._config.test_sections.items()
            if sub.new_in_test(result.test)
        )

    def _evaluate_results(
        self, results: list[LoggedResult], test_num: int
    ) -> tuple[bool, bool, Optional[LoggedResult]]:
        ok, definitive, breaker = check_verdicts(
            list(map(lambda r: r.verdict, results)),
            self._solution.tests[test_num],
        )
        return ok, definitive, results[breaker] if breaker is not None else None

    @staticmethod
    def from_log(
        name: str, config: TaskConfig, testing_log, limit: float
    ) -> "SolutionResults":
        if name not in testing_log["solutions"]:
            raise MissingSolution(name)

        results = []
        for test_name, result in testing_log["solutions"][name]["results"].items():

            def load_decimal(result: dict[str, Any], key: str) -> Optional[Decimal]:
                return Decimal(result[key]) if key in result else None

            results.append(
                limit_result(
                    LoggedResult(
                        Verdict[result["result"]],
                        load_decimal(result, "relative_points"),
                        load_decimal(result, "absolute_points"),
                        result["time"],
                        test_name,
                        Verdict[result["result"]],
                    ),
                    limit,
                )
            )

        return SolutionResults(name, config, results)

    def get_all(self) -> list[LoggedResult]:
        return self._results

    def get_by_test(self) -> list[list[LoggedResult]]:
        by_test: list[list[LoggedResult]] = [
            [] for _ in range(self._config.tests_count)
        ]
        for res in self._results:
            for num, sub in self._config.test_sections.items():
                if sub.in_test(res.test):
                    by_test[num].append(res)

        return by_test

    def check_test(self, num: int) -> Optional[str]:
        results = self.get_by_test()

        expected = self._solution.tests[num]
        ok, _, breaker = self._evaluate_results(results[num], num)
        if not ok:
            failed_test = f" ({breaker.test})" if breaker else ""
            return f"{expected}{failed_test}"
        return None

    def check_points(self) -> Optional[str]:
        achieved = Decimal(0)
        results = self.get_by_test()
        for num, sub in self._config.test_sections.items():
            achieved += min(map(lambda r: r.points(sub.points), results[num]))

        points = self._solution.points
        points_min = self._solution.points_min
        points_max = self._solution.points_max

        if points is not None and points != achieved:
            return f"{points}p"
        if points_min is not None and not achieved >= points_min:
            return f"at least {points_min}p"
        if points_max is not None and not achieved <= points_max:
            return f"at most {points_max}p"

        return None

    def check_all(self) -> list[str]:
        fails = []
        for num in self._config.test_sections:
            if (err := self.check_test(num)) is not None:
                fails.append(f"{err} on {self._config.test_sections[num].name}")
        if (err := self.check_points()) is not None:
            fails.append(f"get {err}")

        return fails

    def get_time_limit_range(self, num: int) -> tuple[float, float]:
        results = self.get_by_test()[num]
        times = [0] + list(map(lambda r: r.time, results))
        times.sort()
        min_possible = len(times) - 1
        max_possible = 0
        for i, time in enumerate(times):
            ok, _, _ = self._evaluate_results(
                list(map(lambda r: limit_result(r, time), results)),
                num,
            )
            if ok:
                min_possible = min(i, min_possible)
                max_possible = max(i, max_possible)

        times.append(inf)
        return (times[min_possible], times[max_possible + 1])


def show_all(results: list[LoggedResult]) -> list[LoggedResult]:
    return results


def show_slowest(results: list[LoggedResult]) -> list[LoggedResult]:
    return [max(results, key=lambda r: (r.verdict.value, r.time))]


def visualize(
    path: str,
    filter: str,
    bundle: bool,
    solutions: Optional[list[str]],
    limit: Optional[float],
    filename: str,
    segments: Optional[int],
    pisek_dir: Optional[str],
    config_filename: str,
    **_,
) -> None:
    config = load_config(path, pisek_dir, config_filename)

    log_path = os.path.join(path, filename)
    try:
        with open(log_path) as log_file:
            testing_log = json.load(log_file)
    except FileNotFoundError:
        raise MissingFile(
            f"File {log_path} not found. Test with --testing-log to create a log."
        )

    if testing_log["source"] == "cms":
        limit_default = config.cms.time_limit
    else:
        limit_default = config.solution_time_limit
    time_limit = limit_default if limit is None else limit

    filter_fn = show_all if filter == "all" else show_slowest

    segment_cnt = terminal_width // 8 if segments is None else segments

    expanded_solutions = expand_solutions(config, solutions)

    results: dict[str, SolutionResults] = {}
    max_test_length = 0
    for sol in expanded_solutions:
        try:
            results[sol] = SolutionResults.from_log(
                sol, config, testing_log, time_limit
            )
            max_test_length = max(
                max_test_length, max(map(lambda r: len(r.test), results[sol].get_all()))
            )
        except MissingSolution as err:
            eprint(ColorSettings.colored(str(err), "yellow"))

    wrong_solutions = {}
    for sol, sol_res in results.items():
        if not bundle:
            sol_err = sol_res.check_points()
            err_msg = ""
            if sol_err is not None:
                err_msg = ColorSettings.colored(f" should get {sol_err}", "red")
                wrong_solutions[sol] = True
            print(f"{sol}{err_msg}")

            for num, group_res in enumerate(sol_res.get_by_test()):
                test_err = sol_res.check_test(num)
                err_msg = ""
                if test_err is not None:
                    err_msg = ColorSettings.colored(f" should result {test_err}", "red")
                    wrong_solutions[sol] = True

                print(tab(f"{config.test_sections[num].name}{err_msg}"))
                for res in filter_fn(group_res):
                    print(
                        tab(tab(res.to_str(time_limit, segment_cnt, max_test_length)))
                    )

        else:
            sol_errs = sol_res.check_all()
            err_msg = ""
            if sol_errs:
                err_msg = ColorSettings.colored(f" should {', '.join(sol_errs)}", "red")
                wrong_solutions[sol] = True
            print(f"{sol}{err_msg}")

            for res in filter_fn(sol_res.get_all()):
                print(tab(tab(res.to_str(time_limit, segment_cnt, max_test_length))))

    print()
    if wrong_solutions:
        print(
            ColorSettings.colored(
                f"Solutions {', '.join(wrong_solutions.keys())} should result differently.",
                "red",
            )
        )

    min_possible = 0.0
    max_possible = inf
    for sol, sol_res in results.items():
        for num in config.test_sections:
            a, b = sol_res.get_time_limit_range(num)
            min_possible = max(a, min_possible)
            max_possible = min(b, max_possible)

    if min_possible <= max_possible:
        limit_msg = f"Valid time limit between {min_possible:.2f}, {max_possible:.2f}."
    else:
        limit_msg = "No valid time limit found."
    print(ColorSettings.colored(limit_msg, "cyan"))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
from dataclasses import dataclass
from enum import Enum
from typing import Optional, Union

from pisek.utils.paths import TaskPath


class RunResultKind(Enum):
    OK = 0
    RUNTIME_ERROR = 1
    TIMEOUT = 2


@dataclass(frozen=True)
class RunResult:
    """Represents the way the program execution ended. Specially, a program
    that finished successfully, but got Wrong Answer, still gets the OK
    RunResult."""

    kind: RunResultKind
    returncode: int
    time: float
    wall_time: float
    stdin_file: TaskPath | int | None = None
    stdout_file: TaskPath | int | None = None
    stderr_file: Optional[TaskPath] = None
    status: str = ""



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from pisek.task_jobs.validator.validator_base import ValidatorJob


class Simple0Validate(ValidatorJob):
    """Runs simple-0 validator on single input."""

    @property
    def _expected_returncode(self):
        return 0

    def _validation_args(self):
        return [str(self.test)]


class Simple42Validate(ValidatorJob):
    """Runs simple-42 validator on single input."""

    @property
    def _expected_returncode(self):
        return 42

    def _validation_args(self):
        return [str(self.test)]



"""
A module for testing pisek itself. The strategy is to take a functioning
fixture of a task and then break it in various small ways to see whether
pisek catches the problem.
"""

import unittest
import os

from util import TestFixtureVariant, overwrite_file, modify_config


class TestSumKasiopea(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "fixtures/sum_kasiopea/"


class TestMissingSampleIn(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "sample.in"))


class TestMissingSampleOut(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "sample.out"))


class TestWrongSampleOut(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        with open(os.path.join(self.task_dir, "sample.out"), "a") as f:
            f.write("0\n")


class TestMissingGenerator(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "gen.cpp"))


class TestBadGenerator(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        generator_filename = os.path.join(self.task_dir, "gen.cpp")
        os.remove(generator_filename)

        with open(generator_filename, "w") as f:
            f.write("int main() { return 0; }\n")


class TestPythonGenerator(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "gen.cpp", "gen_2.py", new_file_name="gen.py")


class TestNonHexaPythonGenerator(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "gen.cpp"))

        new_program = [
            "#!/usr/bin/env python3",
            "import sys",
            "print(sys.argv[1])",
            "print(int(sys.argv[2], 10))",
        ]
        with open(os.path.join(self.task_dir, "gen.py"), "w") as f:
            f.write("\n".join(new_program))


class TestNonHexaGenerator(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "gen.cpp"))

        new_program = [
            "#include <iostream>",
            "#include <string>",
            "int main(int argc, char* argv[]) {",
            "if (argc != 3) { return 1; }",
            "std::cout << argv[1] << std::endl;"
            "std::cout << std::strtoull(argv[2], NULL, 10) << std::endl;"
            "return 0;}",
        ]
        with open(os.path.join(self.task_dir, "gen.cpp"), "w") as f:
            f.write("\n".join(new_program))


class TestDifferentlyScoringSolution(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(
            self.task_dir, "solve_0b.py", "solve_4b.cpp", new_file_name="solve_0b.cpp"
        )


class TestJudge(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge"
            raw_config["tests"]["judge_type"] = "opendata-v2"

        modify_config(self.task_dir, modification_fn)


class TestJudgeWithNoInput(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["judge_needs_in"] = "0"
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge"
            raw_config["tests"]["judge_type"] = "opendata-v2"

        modify_config(self.task_dir, modification_fn)


class TestJudgeWithNoOutput(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["judge_needs_out"] = "0"
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge"
            raw_config["tests"]["judge_type"] = "opendata-v2"

        modify_config(self.task_dir, modification_fn)


class TestPythonJudge(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge_py"
            raw_config["tests"]["judge_type"] = "opendata-v2"

        modify_config(self.task_dir, modification_fn)


class TestBadJudge(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge_bad"
            raw_config["tests"]["judge_type"] = "opendata-v1"

        modify_config(self.task_dir, modification_fn)


class TestPythonCRLF(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        os.remove(os.path.join(self.task_dir, "solve.py"))

        new_program = [
            "#!/usr/bin/env python3",
            "t = int(input())",
            "for i in range(t):",
            "    a, b = [int(x) for x in input().split()]",
            "    c = a + b",
            "    print(c)",
        ]
        with open(os.path.join(self.task_dir, "solve.py"), "w") as f:
            f.write("\r\n".join(new_program))


class TestStrictValidator(TestSumKasiopea):
    """A validator whose bounds are stricter than what the generator creates."""

    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        overwrite_file(self.task_dir, "validate.py", "validate_strict.py")


class TestDirtySample(TestSumKasiopea):
    """Sample without newline at the end."""

    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        sample = ["3", "1 2", "-8 5", "0 0"]
        with open(os.path.join(self.task_dir, "sample.in"), "w") as f:
            f.write("\n".join(sample))


class TestExtraConfigKeys(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def catch_exceptions(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["task"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


class TestExtraConfigKeysInTest(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def catch_exceptions(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["test01"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


class TestExtraConfigSection(TestSumKasiopea):
    def expecting_success(self) -> bool:
        return False

    def catch_exceptions(self) -> bool:
        return True

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config.add_section("baz")
            raw_config["baz"]["foo"] = "bar"

        modify_config(self.task_dir, modification_fn)


if __name__ == "__main__":
    unittest.main()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2025        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from abc import ABC, abstractmethod
from colorama import Cursor, ansi
from functools import wraps
from math import ceil
import re
import sys
import time
from typing import Callable, Concatenate, ParamSpec

from pisek.utils.text import tab
from pisek.utils.terminal import terminal_width, terminal_height, LINE_SEPARATOR

from pisek.jobs.jobs import State, PipelineItem, Job, JobManager
from pisek.env.env import Env

P = ParamSpec("P")


class Reporter(ABC):
    def __init__(self, env: Env, job_managers: list[JobManager]) -> None:
        self._env = env
        self._job_managers = job_managers

    @abstractmethod
    def update(self, active_jobs: list[Job]) -> None:
        pass

    @abstractmethod
    def report_job(self, job: Job) -> None:
        pass

    @abstractmethod
    def report_manager(self, job_manager: JobManager) -> None:
        pass


class CommandLineReporter(Reporter):
    def __init__(self, env: Env, job_managers: list[JobManager]) -> None:
        self._finished_jm = 0
        self._tmp_lines = 0
        self._dirty_lines = 0
        self._last_update = time.time()
        super().__init__(env, job_managers)

    def update(self, active_jobs: list[Job]) -> None:
        self._reset_tmp_lines()
        for i in range(self._finished_jm, len(self._job_managers)):
            job_man = self._job_managers[i]

            if i == self._finished_jm and job_man.state.finished():
                self._finished_jm += 1
                self._print(job_man.get_status())
                if job_man.state != State.cancelled:
                    self._report_manager(job_man)
                if job_man.state == State.failed and not self._env.full:
                    return
            elif job_man.state == State.running:
                msg = job_man.get_status()
                self._print_tmp(msg)
                break

        now = time.time()
        if terminal_height <= self._env.jobs + 5:  # Some extra safety margin
            if active_jobs:
                self._print_tmp(
                    f"{len(active_jobs)} active job{'s' if len(active_jobs) >= 2 else ''}, longest running: {self._format_job(active_jobs[0], now)}"
                )
        else:
            if active_jobs:
                self._print_tmp("Active jobs:")

            for job in active_jobs:
                self._print_tmp("- " + self._format_job(job, now))

            self._clear_lines(self._dirty_lines)

    def report_job(self, job: Job) -> None:
        pass

    def report_manager(self, job_manager: JobManager) -> None:
        pass

    def _report_manager(self, job_manager: JobManager) -> None:
        self._clear_lines(self._dirty_lines)
        report_about: list[PipelineItem] = job_manager.jobs + [job_manager]

        # Prints and warnings
        for item in report_about:
            for msg, is_stderr in item.terminal_output:
                self._print(msg, end="", file=sys.stderr if is_stderr else sys.stdout)

        # Fails
        fails: list[str] = []
        for item in report_about:
            if item.state == State.failed and item.fail_msg:
                fails.append(self._fail_message(item))

        if fails:
            msg = self._colored(
                LINE_SEPARATOR + LINE_SEPARATOR.join(fails) + LINE_SEPARATOR, "red"
            )
            self._print(msg, end="")

    def _format_job(self, job: Job, now: float) -> str:
        run_time: float = 0 if job.started is None else max(0, now - job.started)
        return f"{job.name} ({run_time:.1f}s)"

    @staticmethod
    def _jumps(
        func: Callable[Concatenate["CommandLineReporter", P], None],
    ) -> Callable[Concatenate["CommandLineReporter", P], None]:
        @wraps(func)
        def g(self, *args: P.args, **kwargs: P.kwargs) -> None:
            if not self._env.no_jumps:
                func(self, *args, **kwargs)

        return g

    @_jumps
    def _reset_tmp_lines(self) -> None:
        print(Cursor.UP() * self._tmp_lines, end="")
        self._dirty_lines += self._tmp_lines
        self._tmp_lines = 0

    @_jumps
    def _print_tmp(self, msg: str, *args, **kwargs) -> None:
        """Prints a text to be rewritten latter."""
        lines = self._lines(msg)
        self._clear_lines(lines)
        self._tmp_lines += lines
        self._dirty_lines -= lines
        print(msg, *args, **kwargs)

    def _print(self, msg: str, *args, **kwargs) -> None:
        """Prints a text."""
        self._reset_tmp_lines()
        self._clear_lines(self._lines(msg))
        self._dirty_lines -= self._lines(msg)
        print(msg, *args, **kwargs)

    @_jumps
    def _clear_lines(self, count: int) -> None:
        if count >= self._dirty_lines:
            self._dirty_lines = 0

        # Clearing more than terminal height just makes empty lines
        count = min(count, terminal_height)

        # We don't want to print any newlines when count == 1, as it makes flashes
        print(
            "\n" * (count - 1)
            + f"{ansi.clear_line()}{Cursor.UP()}" * (count - 1)
            + ansi.clear_line(),
            end="",
        )

    def _lines(self, text: str) -> int:
        ansi_escape = re.compile(r"\x1B(?:[@-Z\\-_]|\[[0-?]*[ -/]*[@-~])")
        return sum(
            max(ceil(len(re.sub(ansi_escape, "", line)) / terminal_width), 1)
            for line in text.split("\n")
        )

    def _colored(self, msg: str, color: str) -> str:
        return self._env.colored(msg, color)

    def _fail_message(self, pitem: PipelineItem) -> str:
        return f'"{pitem.name}" failed:\n{tab(pitem.fail_msg)}\n'



import configparser
import io
import os
import shutil
import tempfile
from typing import Callable
import unittest
from unittest import mock

from pisek.user_errors import UserError
from pisek.config import config_hierarchy
from pisek.__main__ import test_task_path
from pisek.utils.util import clean_task_dir
from pisek.utils.pipeline_tools import assert_task_dir


class TestFixture(unittest.TestCase):
    @property
    def fixture_path(self) -> str | None:
        return None

    def set_env(self) -> None:
        pass

    def setUp(self) -> None:
        os.environ["LOG_LEVEL"] = "debug"

        if not self.fixture_path:
            return
        assert self.fixture_path[-1] == "/"

        self.task_dir_orig = os.path.abspath(
            os.path.join(os.path.dirname(__file__), "..", self.fixture_path)
        )
        self.fixtures_dir = tempfile.mkdtemp(prefix="pisek-test_")
        self.task_dir = os.path.join(
            self.fixtures_dir, os.path.basename(os.path.dirname(self.fixture_path))
        )

        # shutil.copytree() requires that the destination directory does not exist,
        os.rmdir(self.fixtures_dir)
        shutil.copytree(
            self.task_dir_orig,
            self.task_dir,
        )

        pisek_dir = os.path.join(self.task_dir_orig, "../pisek")
        if os.path.exists(pisek_dir):
            os.environ["PISEK_DIRECTORY"] = "../pisek"
            shutil.copytree(
                pisek_dir,
                os.path.join(self.fixtures_dir, "pisek"),
            )

        self.set_env()

        assert_task_dir(
            self.task_dir,
            os.environ.get("PISEK_DIRECTORY"),
            config_hierarchy.DEFAULT_CONFIG_FILENAME,
        )
        clean_task_dir(self.task_dir)

        self.cwd_orig = os.getcwd()
        os.chdir(self.task_dir)

    def runTest(self) -> None:
        # Implement this!
        pass

    def tearDown(self) -> None:
        if not self.fixture_path:
            return

        os.chdir(self.cwd_orig)

        assert self.fixtures_dir.startswith("/tmp") or self.fixtures_dir.startswith(
            "/var"
        )
        shutil.rmtree(self.fixtures_dir)

    def log_files(self) -> None:
        """Log all files for checking whether new ones have been created."""
        self.original_files = os.listdir(self.task_dir)

    def created_files(self) -> list[str]:
        """Additional files that are expected to be created."""
        return []

    def check_files(self) -> None:
        """
        Check whether there are no new unexpected files.
        Ignored:
            .pisek_cache data/* build/*
        """
        directories = ["build", "tests", ".pisek"]

        all_paths = set(self.original_files + directories + self.created_files())

        for path in os.listdir(self.task_dir):
            self.assertIn(
                path,
                all_paths,
                f"Pisek generated new file {path}.",
            )


class TestFixtureVariant(TestFixture):
    def expecting_success(self) -> bool:
        return True

    def catch_exceptions(self) -> bool:
        return False

    def modify_task(self) -> None:
        """
        Code which modifies the task before running the tests should go here.
        For example, if we want to check that the presence of `sample.in` is
        correctly checked for, we would remove the file here.
        """
        pass

    @property
    def time_limit(self) -> float | None:
        return None

    def runTest(self) -> None:
        if not self.fixture_path:
            return

        self.set_env()
        self.modify_task()
        self.log_files()

        # We lower the time limit to make the self-tests run faster. The solutions
        # run instantly, with the exception of `solve_slow_4b`, which takes 10 seconds
        # and we want to consider it a time limit
        @mock.patch("sys.stdout", new_callable=io.StringIO)
        @mock.patch("sys.stderr", new_callable=io.StringIO)
        def run(*args) -> bool:
            try:
                test_task_path(
                    self.task_dir,
                    inputs=1,
                    strict=False,
                    full=False,
                    time_limit=self.time_limit,
                    plain=False,
                    pisek_dir=os.environ.get("PISEK_DIRECTORY"),
                )
                return True
            except UserError:
                return False

        self.assertEqual(run(), self.expecting_success())

        self.check_end_state()
        self.check_files()

    def check_end_state(self):
        # Here we can verify whether some conditions hold when Pisek finishes,
        # making sure that the end state is reasonable
        pass


def overwrite_file(
    task_dir: str, old_file: str, new_file: str, new_file_name: str | None = None
) -> None:
    os.remove(os.path.join(task_dir, old_file))
    shutil.copy(
        os.path.join(task_dir, new_file),
        os.path.join(task_dir, new_file_name or old_file),
    )


def modify_config(
    task_dir: str, modification_fn: Callable[[configparser.ConfigParser], None]
) -> None:
    """
    `modification_fn` accepts the config (in "raw" ConfigParser format) and may
    modify it. The modified version is then saved.

    For example, if we want to change the evaluation method ("out_check")
    from `diff` to `judge`, we would do that in `modification_fn` via:
        config["tests"]["out_check"] = "judge"
        config["tests"]["out_judge"] = "judge"  # To specify the judge program file
    """

    config = config_hierarchy.new_config_parser()
    config_path = os.path.join(task_dir, config_hierarchy.DEFAULT_CONFIG_FILENAME)
    read_files = config.read(config_path)
    if not read_files:
        raise FileNotFoundError(f"Missing configuration file {config_path}.")

    modification_fn(config)

    with open(config_path, "w") as f:
        config.write(f)



#!/usr/bin/env python3
import sys
from typing import NoReturn


def accept() -> NoReturn:
    print(1)
    print("OK", file=sys.stderr)
    quit()


def reject() -> NoReturn:
    print(0)
    print("OK", file=sys.stderr)
    quit()


def solve(file) -> int:
    n = int(file.readline())
    vals = list(map(int, file.readline().split()))
    assert n == len(vals)
    return max(vals)


if len(sys.argv) == 1:
    print(solve(sys.stdin))
else:
    with open(sys.argv[3]) as f:
        try:
            res = int(f.read())
        except ValueError:
            reject()

    with open(sys.argv[1]) as f:
        if res == solve(f):
            accept()
        else:
            reject()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
from typing import Optional
import uuid

from pisek.utils.text import tab
from pisek.utils.paths import TaskPath, BUILD_DIR

from pisek.env.env import Env, TestingTarget
from pisek.config.task_config import BuildSection, RunSection
from pisek.config.config_types import BuildStrategyName, OutCheck

from pisek.task_jobs.tools import PrepareTokenJudge, PrepareShuffleJudge
from pisek.jobs.jobs import Job, PipelineItemFailure
from pisek.task_jobs.task_job import TaskJob
from pisek.task_jobs.task_manager import TaskJobManager

from pisek.task_jobs.builder.strategies import (
    BuildStrategy,
    AUTO_STRATEGIES,
    ALL_STRATEGIES,
)

WORKING_DIR_BASE = "_workspace"


class BuildManager(TaskJobManager):
    """Builds task programs."""

    def __init__(self):
        super().__init__("Build programs")

    def _build_program_job(self, run: Optional[RunSection]) -> Optional["Build"]:
        if run is None or run.build.section_name in self._built_sections:
            return None
        self._built_sections.add(run.build.section_name)
        return Build(self._env, run.build)

    def _get_jobs(self) -> list[Job]:
        self._built_sections: set[str] = set()
        jobs: list[Job | None] = []

        jobs.append(self._build_program_job(self._env.config.tests.in_gen))
        jobs.append(self._build_program_job(self._env.config.tests.validator))

        # If builds from previous run ended unsuccessfully, clean up
        for path in os.listdir(BUILD_DIR):
            if path.startswith(WORKING_DIR_BASE):
                shutil.rmtree(os.path.join(BUILD_DIR, path))

        if self._env.target in (
            TestingTarget.build,
            TestingTarget.solutions,
            TestingTarget.all,
        ):
            if self._env.config.tests.out_check == OutCheck.judge:
                jobs.append(self._build_program_job(self._env.config.tests.out_judge))
            elif self._env.config.tests.out_check == OutCheck.tokens:
                jobs.append(PrepareTokenJudge(self._env))
            elif self._env.config.tests.out_check == OutCheck.shuffle:
                jobs.append(PrepareShuffleJudge(self._env))

            for solution in self._env.solutions:
                jobs.append(
                    self._build_program_job(self._env.config.solutions[solution].run)
                )

        filtered_jobs: list[Job] = []
        for j in jobs:
            if j is not None:
                filtered_jobs.append(j)

        return filtered_jobs


class Build(TaskJob):
    """Job that compiles a program."""

    def __init__(
        self,
        env: Env,
        build_section: BuildSection,
        **kwargs,
    ) -> None:
        super().__init__(env=env, name=f"Build {build_section.program_name}", **kwargs)
        self.build_section = build_section

    def _resolve_program(self, glob: TaskPath) -> set[TaskPath]:
        result = self._globs_to_files([f"{glob.path}.*", glob.path], TaskPath("."))
        if len(result) == 0:
            raise PipelineItemFailure(f"No paths found for {glob.col(self._env)}.")
        return set(result)

    def _check_valid_sources(self, sources: set[TaskPath]) -> None:
        """Checks that sources are one directory or multiple files."""
        source_dir = any(map(self._is_dir, sources))
        source_file = any(map(self._is_file, sources))
        if source_dir and source_file:
            raise PipelineItemFailure(
                f"Mixed files and directories for sources:\n"
                + tab(self._path_list(list(sorted(sources))))
            )
        elif source_dir and len(sources) > 1:
            raise PipelineItemFailure(
                f"Only one directory allowed in sources:\n"
                + tab(self._path_list(list(sorted(sources))))
            )

    def _strategy_sources(
        self, strategy: type[BuildStrategy], sources: set[TaskPath]
    ) -> set[TaskPath]:
        new_sources = set()
        if strategy.extra_sources is not None:
            for part in getattr(self.build_section, strategy.extra_sources):
                new_sources |= self._resolve_program(part)
        return sources | new_sources

    def _strategy_extras(
        self, strategy: type[BuildStrategy], extras: set[TaskPath]
    ) -> set[TaskPath]:
        new_extras = set()
        if strategy.extra_nonsources is not None:
            new_extras = set(getattr(self.build_section, strategy.extra_nonsources))
        return extras | new_extras

    def _run(self) -> None:
        sources: set[TaskPath] = set()
        extras: set[TaskPath] = set(self.build_section.extras)
        for part in self.build_section.sources:
            sources |= self._resolve_program(part)

        # We need to check valid sources twice:
        # - First in order to report correct error
        # - Second time because we added extra sources
        self._check_valid_sources(sources)

        if self.build_section.strategy == BuildStrategyName.auto:
            strategy_cls = self._resolve_strategy(sources)
        else:
            strategy_cls = ALL_STRATEGIES[self.build_section.strategy]

        sources = self._strategy_sources(strategy_cls, sources)
        extras = self._strategy_extras(strategy_cls, extras)
        self._check_valid_sources(sources)

        workdir = os.path.join(BUILD_DIR, f"{WORKING_DIR_BASE}_{uuid.uuid4()}")
        os.makedirs(workdir)

        for path in sources | extras:
            # Intentionally avoiding caching results
            dst = os.path.join(workdir, path.name)
            if os.path.exists(dst):
                raise PipelineItemFailure(
                    f"Duplicate filename / dirname: '{path.name}'"
                )
            if self._is_dir(path):
                shutil.copytree(path.path, dst)
                self._access_dir(path)
            elif self._is_file(path):
                shutil.copy(path.path, dst)
                self._access_file(path)
            else:
                raise PipelineItemFailure(f"No path {path.col(self._env)} exists.")

        target = TaskPath(BUILD_DIR, self.build_section.program_name)
        self.make_filedirs(target)
        if os.path.isdir(target.path):
            shutil.rmtree(target.path)
        elif os.path.isfile(target.path):
            os.remove(target.path)

        strategy = strategy_cls(self.build_section, self._env, self._run_subprocess)
        executable_name = strategy.build(
            workdir,
            list(map(lambda p: p.name, sources)),
            list(map(lambda p: p.name, extras)),
        )

        if self._env.verbosity >= 1 or strategy.stderr_output:
            msg = f"Built '{self.build_section.program_name}' using build strategy '{strategy_cls.name}'."
            self._print(self._colored(msg, "magenta"))
            self._print(strategy.stderr_output, end="", stderr=True)

        executable = os.path.join(workdir, executable_name)
        # Intentionally avoiding caching sources
        if os.path.isdir(executable):
            shutil.copytree(executable, target.path, symlinks=True)
            self._access_dir(target)
        else:
            shutil.copy(executable, target.path)
            self._access_file(target)

        shutil.rmtree(workdir)

    def _resolve_strategy(self, sources: set[TaskPath]) -> type[BuildStrategy]:
        applicable = []
        for strategy in AUTO_STRATEGIES:
            strat_sources = self._strategy_sources(strategy, sources)
            if strategy.applicable(
                self.build_section, list(map(lambda p: p.path, strat_sources))
            ):
                applicable.append(strategy)
        if len(applicable) == 0:
            raise PipelineItemFailure(
                f"No applicable build strategy for [{self.build_section.section_name}] with sources:\n"
                + tab(self._path_list(list(sorted(sources))))
            )
        elif len(applicable) >= 2:
            names = " ".join(s.name for s in applicable)
            raise RuntimeError(f"Multiple strategies applicable: {names}")

        return applicable[0]



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from enum import StrEnum, auto
import os
from pydantic import Field
from typing import Optional

from pisek.utils.colors import ColorSettings
from pisek.env.base_env import BaseEnv
from pisek.config.config_hierarchy import DEFAULT_CONFIG_FILENAME
from pisek.config.task_config import load_config, TaskConfig
from pisek.config.select_solutions import expand_solutions


class TestingTarget(StrEnum):
    all = auto()
    build = auto()
    generator = auto()
    solutions = auto()


class Env(BaseEnv):
    """
    Collection of environment variables for task testing.

    Attributes:
        target: What is being tested
        config: Environment variables defined by task config
        jobs: How many jobs to run at most in parallel
        verbosity: How much verbose to be
        file_contents: Show file contents in errors
        full: Whether to stop after the first failure
        no_colors: If not to use ansi colors
        no_jumps: If not to use ansi control sequences
        strict: Whether to interpret warnings as failures
        testing_log: Whether to produce testing_log.json after running
        solutions: List of all solutions to be tested
        time_limit: Time limit for solutions in seconds. Overrides task config if specified. (Must be >= 0)
        all_inputs: Finish testing all inputs of a solution
        repeat: Test task REPEAT times giving generator different seeds. (Changes seeded inputs only.)
        iteration: Current iteration of task testing. (0 <= iteration < repeat)
    """

    target: TestingTarget
    config: TaskConfig
    jobs: int
    verbosity: int
    file_contents: bool
    full: bool
    no_colors: bool
    no_jumps: bool
    strict: bool
    testing_log: bool
    solutions: list[str]
    time_limit: Optional[float] = Field(ge=0)
    all_inputs: bool
    repeat: int = Field(ge=1)
    iteration: int = Field(ge=0)

    @staticmethod
    def load(
        target: str = TestingTarget.all,
        jobs: int | None = None,
        verbosity: int = 0,
        file_contents: bool = False,
        full: bool = False,
        all_inputs: bool = False,
        plain: bool = False,
        no_jumps: bool = False,
        no_colors: bool = False,
        strict: bool = False,
        testing_log: bool = False,
        solutions: Optional[list[str]] = None,
        time_limit: Optional[float] = None,
        repeat: int = 1,
        iteration: int = 0,
        pisek_dir: Optional[str] = None,
        config_filename: str = DEFAULT_CONFIG_FILENAME,
        **_,
    ) -> "Env":
        no_jumps |= plain
        no_colors |= plain

        config = load_config(".", pisek_dir, config_filename, strict)

        if target == TestingTarget.build:
            assert solutions is None
            solutions = [config.primary_solution]

        expanded_solutions = expand_solutions(config, solutions)

        if expanded_solutions and config.tests.judge_needs_out:
            if config.primary_solution in expanded_solutions:
                expanded_solutions.remove(config.primary_solution)
            expanded_solutions.insert(0, config.primary_solution)

        return Env(
            target=TestingTarget(TestingTarget.all if target is None else target),
            jobs=(
                max(1, min((os.cpu_count() or 2) // 2, 32)) if jobs is None else jobs
            ),
            config=config,
            verbosity=verbosity,
            file_contents=file_contents,
            full=full,
            no_jumps=no_jumps,
            no_colors=no_colors,
            strict=strict,
            testing_log=testing_log,
            solutions=expanded_solutions,
            time_limit=time_limit,
            all_inputs=all_inputs,
            repeat=repeat,
            iteration=iteration,
        )

    def colored(self, msg: str, color: str) -> str:
        self.no_colors  # Caching
        return ColorSettings.colored(msg, color)



from pisek.jobs.jobs import Job
from pisek.jobs.status import StatusJobManager
from pisek.jobs.job_pipeline import JobPipeline
from pisek.utils.paths import IInputPath, IOutputPath, IRawPath

from pisek.task_jobs.tools import sanitize_job, sanitize_job_direct
from pisek.task_jobs.data.testcase_info import TestcaseInfo, TestcaseGenerationMode
from pisek.task_jobs.generator.generator_manager import generate_input_direct
from pisek.task_jobs.solution.solution import RunBatchSolution
from pisek.task_jobs.checker.checker import checker_job
from pisek.opendata.types import OpendataVerdict


class OpendataPipeline(JobPipeline):
    def __init__(
        self,
        gen_input: bool,
        gen_output: bool,
        check: bool,
        input_: IInputPath,
        info: TestcaseInfo,
        test: int,
        seed: int | None,
        correct_output: IOutputPath,
        contestant_output: IRawPath | None,
    ):
        super().__init__()
        self.job_managers = []

        if gen_input:
            self.job_managers.append(InputManager(input_, info, seed))

        if gen_output:
            self.job_managers.append(OutputManager(input_, info, correct_output))

        self._checker_man: CheckerManager | None = None
        if check:
            assert contestant_output is not None
            self._checker_man = CheckerManager(
                input_, test, seed, correct_output, contestant_output
            )
            self.job_managers.append(self._checker_man)

        for i in range(1, len(self.job_managers)):
            self.job_managers[i].add_prerequisite(self.job_managers[i - 1])

    @property
    def verdict(self) -> OpendataVerdict:
        assert self._checker_man is not None
        return self._checker_man.judging_result


class InputManager(StatusJobManager):
    def __init__(self, input_: IInputPath, info: TestcaseInfo, seed: int | None):
        super().__init__(f"Generate input {input_:n}")
        self._input = input_
        self._info = info
        self._seed = seed

    def _get_jobs(self) -> list[Job]:
        if self._info.generation_mode == TestcaseGenerationMode.generated:
            jobs: list[Job] = [
                gen := generate_input_direct(
                    self._env, self._info, self._seed, self._input
                ),
            ]
        else:
            raise NotImplementedError()

        sanitize = sanitize_job(self._env, self._input, True)
        if sanitize is not None:
            jobs.append(sanitize)
            sanitize.add_prerequisite(gen)

        return jobs


class OutputManager(StatusJobManager):
    def __init__(self, input_: IInputPath, info: TestcaseInfo, output: IOutputPath):
        self._input = input_
        self._info = info
        self._output = output
        super().__init__(f"Generate output {self._output:n}")

    def _get_jobs(self) -> list[Job]:
        if self._info.generation_mode == TestcaseGenerationMode.static:
            raise NotImplementedError()
        else:
            jobs: list[Job] = [
                solve := RunBatchSolution(
                    self._env,
                    self._env.config.solutions[self._env.config.primary_solution].run,
                    True,
                    self._input,
                    self._output,
                ),
            ]

        sanitize = sanitize_job(self._env, self._output, False)
        if sanitize is not None:
            jobs.append(sanitize)
            sanitize.add_prerequisite(solve, name="create-source")

        return jobs


class CheckerManager(StatusJobManager):
    def __init__(
        self,
        input_: IInputPath,
        test: int,
        seed: int | None,
        correct_output: IOutputPath,
        contestant_output: IRawPath,
    ):
        self._input = input_
        self._test = test
        self._seed = seed
        self._correct_output = correct_output
        self._contestant_output = contestant_output
        super().__init__(f"Check {contestant_output:n}")

    def _get_jobs(self) -> list[Job]:
        jobs: list[Job] = []

        sanitize = sanitize_job_direct(
            self._env,
            self._contestant_output,
            self._contestant_output.to_sanitized_output(),
            False,
        )
        self._check = checker_job(
            self._input,
            self._contestant_output.to_sanitized_output(),
            self._correct_output,
            self._test,
            self._seed,
            None,
            self._env,
        )

        if sanitize is not None:
            jobs.append(sanitize)
            self._check.add_prerequisite(sanitize, name="sanitize")
        jobs.append(self._check)

        return jobs

    @property
    def judging_result(self) -> OpendataVerdict:
        res = self._check.result
        assert res is not None
        return OpendataVerdict(
            res.verdict,
            res.message,
            res.points(self._env, self._env.config.test_sections[self._test].points),
            res.log,
            res.note,
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2024        Antonín Maloň  <git@tonyl.eu>
# Copyright (c)   2024        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, NoReturn, Optional

from pisek.utils.text import tab
from pisek.env.env import Env
from pisek.config.config_types import ProgramRole
from pisek.config.task_config import RunSection
from pisek.utils.paths import TaskPath, IInputPath, LogPath
from pisek.task_jobs.program import ProgramsJob, RunResultKind
from pisek.task_jobs.data.testcase_info import TestcaseInfo

from .base_classes import GeneratorListInputs, GenerateInput, GeneratorTestDeterminism


class PisekV1ListInputs(GeneratorListInputs):
    """Lists all inputs for pisek-v1 generator."""

    def __init__(self, env: Env, generator: RunSection, **kwargs) -> None:
        super().__init__(env=env, generator=generator, **kwargs)

    def _run(self) -> list[TestcaseInfo]:
        testcase_infos = []
        input_names: set[str] = set()
        self._create_inputs_list()
        for i, line in enumerate(self._get_input_lines()):
            testcase_info = self._get_testcase_info_from_line(line, i)
            if testcase_info.name in input_names:
                self._line_invalid(
                    i, line, f"Input '{testcase_info.name}' already listed"
                )
            input_names.add(testcase_info.name)
            testcase_infos.append(testcase_info)
        return testcase_infos

    def _get_testcase_info_from_line(self, line: str, line_index: int) -> TestcaseInfo:
        try:
            return TestcaseInfo.from_str("generated " + line)
        except ValueError as err:
            self._line_invalid(line_index, line, str(err))

    def _line_invalid(self, line_index: int, contents: str, reason: str) -> NoReturn:
        contents = contents.rstrip("\n")
        message = (
            f"Inputs list invalid (line {line_index+1}) - {reason}:\n"
            f"{tab(self._colored(contents, 'yellow'))}\n"
            f"Generator:"
        )
        raise self._create_program_failure(
            message, self._run_result, status=False, stderr=False
        )

    def _create_inputs_list(self) -> None:
        self._run_result = self._run_program(
            ProgramRole.gen,
            self.generator,
            stdout=self._get_inputs_list_path(),
            stderr=LogPath.generator_log(self.generator.name),
        )
        if self._run_result.kind != RunResultKind.OK:
            raise self._create_program_failure(
                f"{self.generator.name} failed to list inputs",
                self._run_result,
                stderr_force_content=True,
            )

    def _get_input_lines(self) -> list[str]:
        with self._open_file(self._get_inputs_list_path()) as f:
            return f.readlines()

    def _get_inputs_list_path(self):
        return TaskPath.data_path("_generated_inputs_list")


class PisekV1GeneratorJob(ProgramsJob):
    """Abstract class for jobs with OnlineGenerator."""

    generator: RunSection
    seed: Optional[int]
    testcase_info: TestcaseInfo
    input_path: IInputPath

    def __init__(self, env: Env, *, name: str = "", **kwargs) -> None:
        super().__init__(env=env, name=name, **kwargs)

    def _gen(self) -> None:
        args = [self.testcase_info.name]

        if self.testcase_info.seeded:
            assert self.seed is not None
            if self.seed < 0:
                raise ValueError(f"seed {self.seed} is negative")
            args.append(f"{self.seed:016x}")

        result = self._run_program(
            ProgramRole.gen,
            self.generator,
            args=args,
            stdout=self.input_path.to_raw(self._env.config.tests.in_format),
            stderr=self.input_path.to_log(self.generator.name),
        )
        if result.kind != RunResultKind.OK:
            raise self._create_program_failure(
                f"{self.generator.name} failed on input {self.testcase_info.name}"
                + (":" if self.seed is None else f", seed {self.seed:016x}:"),
                result,
                stderr_force_content=True,
            )


class PisekV1Generate(PisekV1GeneratorJob, GenerateInput):
    """Generates input with given name."""

    pass


class PisekV1TestDeterminism(PisekV1GeneratorJob, GeneratorTestDeterminism):
    """Tests determinism of generating a given input."""

    pass



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from decimal import Decimal
import filecmp
import glob
from math import ceil
import os
import shutil
from typing import (
    Callable,
    Concatenate,
    Iterable,
    Literal,
    ParamSpec,
    TypeVar,
)

import subprocess
from pisek.utils.text import pad, pad_left
from pisek.utils.terminal import MSG_LEN
from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.utils.text import tab
from pisek.jobs.jobs import State, Job, PipelineItemAbort

T = TypeVar("T")
P = ParamSpec("P")


class TaskHelper:
    _env: Env

    def _globs_to_files(
        self, globs: Iterable[str], directory: TaskPath
    ) -> list[TaskPath]:
        """Get files in given directory that match any glob."""
        files_per_glob = [
            glob.glob(g, root_dir=directory.path, recursive=True, include_hidden=True)
            for g in globs
        ]
        files: list[str] = sum(files_per_glob, start=[])
        files = list(sorted(set(files)))
        return [TaskPath.from_abspath(directory.path, file) for file in files]

    def _format_points(self, points: Decimal | int | None) -> str:
        precision = self._env.config.task.score_precision
        if points is None:
            text = "?" + "." * (precision > 0) + "?" * precision
        else:
            text = f"{points:.{precision}f}"

        return text + "p"

    def _format_time(self, time: float) -> str:
        time_str = f"{time:.2f}"
        if self._env.time_limit is not None:
            tl = self._env.time_limit
        else:
            tl = self._env.config.max_solution_time_limit

        length = len(f"{tl:.2f}")
        return f"{time_str:>{length}}s"

    def _solution_header_verbosity0(
        self, msg: str, points: Decimal | int | None, time: float
    ):
        max_label_len = max(map(len, self._env.config.solutions))
        points_str = pad_left(
            self._format_points(points),
            len(self._format_points(self._env.config.total_points)),
        )
        time_str = self._format_time(time)

        return f"{pad(msg, max(MSG_LEN-1, max_label_len))} {points_str}  {time_str}  "

    def _path_list(self, paths: list[TaskPath]) -> str:
        return "\n".join(path.col(self._env) for path in paths)

    @staticmethod
    def _short_list(arr: list[str], cutoff: int = 1) -> str:
        inputs_text = ", ".join(arr[:cutoff])
        if len(arr) > cutoff:
            inputs_text += ",…"
        return inputs_text

    @staticmethod
    def _short_text(
        text: str,
        style: Literal["h", "t", "ht"] = "h",
        max_lines: int = 10,
        max_chars: int = 100,
    ) -> str:
        """
        Shorten text to max_lines lines and max_chars on line.
        Keep lines from head / tail / both depending on style.
        """
        s_text = []
        for line in text.split("\n"):
            if len(line) > max_chars:
                line = line[: max_chars - 1] + "…"
            s_text.append(line)
        if len(s_text) < max_lines:
            return "\n".join(s_text)

        tail = max(ceil(("t" in style) * max_lines / len(style)) - 1, 0)
        head = max_lines - tail - 1
        return "\n".join(s_text[:head] + ["[…]"] + (s_text[-tail:] if tail else []))

    @staticmethod
    def makedirs(path: TaskPath, exist_ok: bool = True):
        """Make directories"""
        os.makedirs(path.path, exist_ok=exist_ok)

    @staticmethod
    def make_filedirs(path: TaskPath, exist_ok: bool = True):
        """Make directories for given file"""
        dirname = os.path.dirname(path.path)
        if dirname:
            os.makedirs(dirname, exist_ok=exist_ok)


class TaskJob(Job, TaskHelper):
    """Job class that implements useful methods"""

    def _access_dir(self, dirname: TaskPath) -> None:
        for file in self._globs_to_files(["**"], dirname):
            self._access_file(file)

    @staticmethod
    def _file_access(files: int):
        """Adds first i args as accessed files."""

        def dec(
            f: Callable[Concatenate["TaskJob", P], T],
        ) -> Callable[Concatenate["TaskJob", P], T]:
            def g(self: "TaskJob", *args: P.args, **kwargs: P.kwargs) -> T:
                for i in range(files):
                    arg = args[i]
                    assert isinstance(arg, TaskPath)
                    self._access_file(arg)
                return f(self, *args, **kwargs)

            return g

        return dec

    @_file_access(1)
    def _open_file(self, filename: TaskPath, mode="r", **kwargs):
        if "w" in mode:
            self.make_filedirs(filename)
        return open(filename.path, mode, **kwargs)

    @_file_access(1)
    def _read_file(self, filename: TaskPath):
        with self._open_file(filename) as f:
            return f.read()

    @_file_access(1)
    def _exists(self, path: TaskPath) -> bool:
        return os.path.exists(path.path)

    @_file_access(1)
    def _is_file(self, path: TaskPath) -> bool:
        return os.path.isfile(path.path)

    @_file_access(1)
    def _is_dir(self, path: TaskPath) -> bool:
        return os.path.isdir(path.path)

    def _remove_file(self, filename: TaskPath) -> None:
        "Removes given file. It must be created inside this job."
        self._accessed_files.remove(filename.path)
        return os.remove(filename.path)

    @_file_access(1)
    def _file_size(self, filename: TaskPath) -> int:
        return os.path.getsize(filename.path)

    @_file_access(1)
    def _file_not_empty(self, filename: TaskPath) -> bool:
        with self._open_file(filename) as f:
            content = f.read()
        return len(content.strip()) > 0

    @_file_access(2)
    def _copy_file(self, filename: TaskPath, dst: TaskPath) -> None:
        self.make_filedirs(dst)
        shutil.copy(filename.path, dst.path)

    def _copy_dir(self, path: TaskPath, dst: TaskPath) -> None:
        self.make_filedirs(dst)
        shutil.copytree(path.path, dst.path)
        self._access_dir(path)
        self._access_dir(dst)

    def _copy_target(self, path: TaskPath, dst: TaskPath) -> None:
        if self._is_dir(path):
            self._copy_dir(path, dst)
        else:
            self._copy_file(path, dst)

    @_file_access(2)
    def _rename_file(self, filename: TaskPath, dst: TaskPath) -> None:
        self.make_filedirs(dst)
        return os.rename(filename.path, dst.path)

    @_file_access(2)
    def _link_file(
        self, filename: TaskPath, dst: TaskPath, overwrite: bool = False
    ) -> None:
        self.make_filedirs(dst)
        if overwrite and os.path.exists(dst.path):
            os.remove(dst.path)

        # os.link should follow symlinks, but doesn't:
        # https://bugs.python.org/issue37612
        source = filename.path
        while os.path.islink(source):
            source = os.readlink(source)

        try:
            os.link(source, dst.path)
        except OSError:
            shutil.copyfile(source, dst.path)

    @_file_access(2)
    def _symlink_file(
        self, filename: TaskPath, dst: TaskPath, overwrite: bool = False
    ) -> None:
        self.make_filedirs(dst)
        if overwrite and os.path.exists(dst.path):
            os.remove(dst.path)

        return os.symlink(
            os.path.relpath(filename.path, os.path.dirname(dst.path)), dst.path
        )

    @_file_access(2)
    def _files_equal(self, file_a: TaskPath, file_b: TaskPath) -> bool:
        return filecmp.cmp(file_a.path, file_b.path, shallow=False)

    def _globs_to_files(
        self, globs: Iterable[str], directory: TaskPath
    ) -> list[TaskPath]:
        self._accessed_globs |= set(os.path.join(directory.path, g) for g in globs)
        return super()._globs_to_files(globs, directory)

    def _quote_file(self, file: TaskPath, **kwargs) -> str:
        """Get shortened file contents"""
        with self._open_file(file) as f:
            return self._short_text(f.read().strip(), **kwargs)

    def _quote_file_with_name(
        self, file: TaskPath, force_content: bool = False, **kwargs
    ) -> str:
        """
        Get file name (with its shortened content if env.file_contents or force_content)
        for printing into terminal
        """
        if force_content or self._env.file_contents:
            return f"{file.col(self._env)}\n{self._colored(tab(self._quote_file(file, **kwargs)), 'yellow')}\n"
        else:
            return f"{file.col(self._env)}\n"

    def _run_subprocess(self, *args, **kwargs) -> subprocess.Popen:
        process = subprocess.Popen(*args, **kwargs)
        self._wait_for_subprocess(process)
        return process

    def _wait_for_subprocess(self, popen: subprocess.Popen) -> None:
        while popen.poll() is None:
            try:
                popen.wait(timeout=0.2)
            except subprocess.TimeoutExpired:
                pass

            if self.state == State.cancelled:
                popen.terminate()
                raise PipelineItemAbort(self)



#!/usr/bin/env python3

import random
import sys
import os

test_dir = sys.argv[1]
os.makedirs(test_dir, exist_ok=True)

POSITIVE_ONLY = [True, False, False]
MAX_ABS = [int(1e9), int(1e9), int(1e18)]

random.seed(123)

print("Generating...", file=sys.stderr)
for test_i, (positive_only, max_abs) in enumerate(zip(POSITIVE_ONLY, MAX_ABS)):
    for ti in range(5):
        a = random.randint(-max_abs, max_abs)
        b = random.randint(-max_abs, max_abs)
        if positive_only:
            a = abs(a)
            b = abs(b)

        test_filename = "{:0>2}_{:0>2}.in".format(test_i + 1, ti + 1)

        with open(os.path.join(test_dir, test_filename), "w") as f:
            f.write("{} {}\n".format(a, b))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>
# Copyright (c)   2024        Benjamin Swart <benjaminswart@email.cz>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from argparse import Namespace
from datetime import datetime
from functools import wraps
import os
import sys
from typing import Callable, Optional, ParamSpec, TypeVar

from pisek.user_errors import TestingFailed
from pisek.jobs.job_pipeline import JobPipeline
from pisek.utils.util import clean_non_relevant_files, ChangedCWD
from pisek.utils.text import eprint
from pisek.utils.terminal import separator_text
from pisek.utils.paths import INTERNALS_DIR
from pisek.utils.colors import ColorSettings
from pisek.env.env import Env
from pisek.config.task_config import load_config
from pisek.jobs.cache import Cache

P = ParamSpec("P")
R = TypeVar("R")


PATH = "."

LOCK_FILE = os.path.join(INTERNALS_DIR, "lock")


def run_pipeline(
    path: str,
    pipeline_class: Callable[[Env], JobPipeline],
    disable_cache: bool = False,
    **env_args,
) -> None:
    with ChangedCWD(path):
        env = Env.load(**env_args)
        ColorSettings.set_state(not env.no_colors)

        cache = None
        if not disable_cache:
            cache = Cache.load()

        all_accessed_files: set[str] = set()
        for i in range(env.repeat):
            env.iteration = i  # XXX: Dirty trick
            if env.repeat > 1:
                if i != 0:
                    print()
                print(
                    ColorSettings.colored(
                        separator_text(f"Run {i+1}/{env.repeat}"), "cyan"
                    )
                )
                print()

            pipeline = pipeline_class(env.fork())
            failed = pipeline.run_jobs(cache, env)
            if failed:
                raise TestingFailed()
            all_accessed_files |= pipeline.all_accessed_files

        clean_non_relevant_files(all_accessed_files)


class Lock:
    def __init__(self, path):
        self._lock_file = os.path.join(path, LOCK_FILE)
        self._locked = False

    def __enter__(self):
        try:
            os.makedirs(os.path.dirname(self._lock_file), exist_ok=True)
            with open(self._lock_file, "x") as f:
                f.write(f"Locked by pisek at {datetime.now()}")
        except FileExistsError:
            eprint(
                f"Another pisek instance running in same directory. (Lockfile '{LOCK_FILE}')"
            )
            sys.exit(2)

        self._locked = True

    def __exit__(self, exc_type, exc_value, exc_traceback):
        if self._locked and os.path.exists(self._lock_file):
            os.unlink(self._lock_file)


def locked_folder(f: Callable[P, R]) -> Callable[P, R]:
    @wraps(f)
    def g(*args, **kwargs) -> R:
        with Lock(PATH):
            res = f(*args, **kwargs)
        return res

    return g


def with_env(fun: Callable[[Env, Namespace], None]) -> Callable[[Namespace], None]:
    @wraps(fun)
    def wrap(args: Namespace) -> None:
        env = Env.load(**vars(args))
        return fun(env, args)

    return wrap


def assert_task_dir(
    task_dir: str, pisek_directory: Optional[str], config_filename: str
) -> None:
    # XXX: Safeguard, raises an exception if task_dir isn't really a task
    # directory
    load_config(task_dir, pisek_directory, config_filename, suppress_warnings=True)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import os
import shutil
import logging
from typing import Optional, Union

from pisek.utils.paths import BUILD_DIR, TESTS_DIR, INTERNALS_DIR
from pisek.utils.text import fatal_user_error


class ChangedCWD:
    def __init__(self, path):
        self._path = path

    def __enter__(self):
        self._orig_path = os.getcwd()
        os.chdir(self._path)

    def __exit__(self, exc_type, exc_value, exc_traceback):
        os.chdir(self._orig_path)


def clean_task_dir(task_dir: str) -> None:
    for subdir in [BUILD_DIR, TESTS_DIR, INTERNALS_DIR]:
        full = os.path.join(task_dir, subdir)
        try:
            shutil.rmtree(full)
        except FileNotFoundError:
            pass


def clean_non_relevant_files(accessed_files: set[str]) -> None:
    accessed_dirs = {os.path.dirname(file) for file in accessed_files}
    for root, _, files in os.walk(TESTS_DIR):
        for file in files:
            path = os.path.join(root, file)
            if root in accessed_dirs and path not in accessed_files:
                os.remove(path)


def log_level_mapper(log_level: str) -> int:
    log_level = log_level.lower()
    if log_level == "debug":
        return logging.DEBUG
    elif log_level == "info":
        return logging.INFO
    elif log_level == "warning":
        return logging.WARNING
    elif log_level == "error":
        return logging.ERROR
    elif log_level == "critical":
        return logging.CRITICAL
    else:
        fatal_user_error(
            f"Invalid log level - {log_level}!, Please set a valid log level."
        )



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import cast, Any, Optional
from hashlib import blake2b

from pisek.env.env import Env
from pisek.utils.paths import IInputPath, IOutputPath
from pisek.config.config_types import GenType
from pisek.config.task_config import RunSection
from pisek.jobs.jobs import Job, JobManager
from pisek.task_jobs.task_manager import TaskJobManager
from pisek.task_jobs.data.data import InputSmall, OutputSmall
from pisek.task_jobs.tools import sanitize_job
from pisek.task_jobs.validator.validator_base import ValidatorJob
from pisek.task_jobs.validator.validators import VALIDATORS
from pisek.task_jobs.data.testcase_info import TestcaseInfo, TestcaseGenerationMode

from .base_classes import (
    GeneratorListInputs,
    GenerateInput,
    GeneratorTestDeterminism,
    GeneratorRespectsSeed,
)
from .cms_old import CmsOldListInputs, CmsOldGenerate
from .opendata_v1 import (
    OpendataV1ListInputs,
    OpendataV1Generate,
    OpendataV1TestDeterminism,
)
from .pisek_v1 import (
    PisekV1ListInputs,
    PisekV1Generate,
    PisekV1TestDeterminism,
)

SEED_BYTES = 8
SEED_RANGE = range(0, 1 << (SEED_BYTES * 8))


class PrepareGenerator(TaskJobManager):
    """Prepares generator for use."""

    def __init__(self):
        self._inputs = []
        super().__init__("Prepare generator")

    def _get_jobs(self) -> list[Job]:
        assert self._env.config.tests.in_gen is not None
        jobs: list[Job] = [
            list_inputs := list_inputs_job(self._env, self._env.config.tests.in_gen),
        ]
        self._list_inputs = list_inputs

        return jobs

    def _compute_result(self) -> dict[str, Any]:
        return {"inputs": self._list_inputs.result}


def list_inputs_job(env: Env, generator: RunSection) -> GeneratorListInputs:
    LIST_INPUTS: dict[GenType, type[GeneratorListInputs]] = {
        GenType.opendata_v1: OpendataV1ListInputs,
        GenType.cms_old: CmsOldListInputs,
        GenType.pisek_v1: PisekV1ListInputs,
    }

    assert env.config.tests.gen_type is not None
    return LIST_INPUTS[env.config.tests.gen_type](env=env, generator=generator)


def generate_input(
    env: Env, testcase_info: TestcaseInfo, seed: Optional[int]
) -> GenerateInput:
    return generate_input_direct(
        env, testcase_info, seed, testcase_info.input_path(seed)
    )


def generate_input_direct(
    env: Env, testcase_info: TestcaseInfo, seed: Optional[int], input_path: IInputPath
) -> GenerateInput:
    assert env.config.tests.in_gen is not None
    assert env.config.tests.gen_type is not None
    return {
        GenType.opendata_v1: OpendataV1Generate,
        GenType.cms_old: CmsOldGenerate,
        GenType.pisek_v1: PisekV1Generate,
    }[env.config.tests.gen_type](
        env=env,
        generator=env.config.tests.in_gen,
        testcase_info=testcase_info,
        seed=seed,
        input_path=input_path,
    )


def generator_test_determinism(
    env: Env, generator: RunSection, testcase_info: TestcaseInfo, seed: Optional[int]
) -> Optional[GeneratorTestDeterminism]:
    TEST_DETERMINISM = {
        GenType.opendata_v1: OpendataV1TestDeterminism,
        GenType.pisek_v1: PisekV1TestDeterminism,
    }

    if env.config.tests.gen_type not in TEST_DETERMINISM:
        return None
    return TEST_DETERMINISM[env.config.tests.gen_type](
        env=env, generator=generator, testcase_info=testcase_info, seed=seed
    )


class TestcaseInfoMixin(JobManager):
    def __init__(self, name: str, **kwargs) -> None:
        self.inputs: dict[str, tuple[set[int], int | None]] = {}
        self.input_dataset: set[IInputPath] = set()
        self._gen_inputs_job: dict[Optional[int], GenerateInput] = {}

        self._jobs: list[Job] = []
        self._testcase_last: Job | None = None

        super().__init__(name=name, **kwargs)

    def _add_job(
        self,
        job: Job | None,
        new_last: bool = False,
        prerequisite_name: str | None = None,
    ) -> None:
        if job is not None:
            self._jobs.append(job)
            job.add_prerequisite(self._testcase_last, prerequisite_name)
            if new_last:
                self._testcase_last = job

    def _begin_new_testcase(self) -> None:
        self._testcase_last = None

    def _get_seed(self, iteration: int, testcase_info: TestcaseInfo) -> int:
        name_hash = blake2b(digest_size=SEED_BYTES)
        name_hash.update(
            f"{self._env.iteration} {iteration} {testcase_info.name}".encode()
        )
        return int.from_bytes(name_hash.digest())

    def _add_testcase_info_jobs(self, testcase_info: TestcaseInfo, test: int) -> None:
        seeds: list[Optional[int]]
        if testcase_info.seeded:
            seeds = []
            for i in range(testcase_info.repeat):
                seeds.append(self._get_seed(i, testcase_info))
        else:
            seeds = [None]

        self._gen_inputs_job = {}

        skipped: bool = False
        for i, seed in enumerate(seeds):
            if self._skip_testcase(testcase_info, seed, test):
                skipped = True
                self._register_skipped_testcase(testcase_info, seed, test)
                continue

            self._begin_new_testcase()
            input_path = testcase_info.input_path(seed)
            self.inputs[input_path.name] = ({test}, seed)
            self.input_dataset.add(input_path)

            self._add_generate_input_jobs(testcase_info, seed, test, i == 0)
            self._add_solution_jobs(testcase_info, seed, test)

        if (
            self._env.config.checks.generator_respects_seed
            and testcase_info.seeded
            and not skipped
        ):
            self._add_respects_seed_jobs(testcase_info, cast(list[int], seeds), test)

    def _skip_testcase(
        self, testcase_info: TestcaseInfo, seed: Optional[int], test: int
    ) -> bool:
        return testcase_info.input_path(seed) in self.input_dataset

    def _register_skipped_testcase(
        self, testcase_info: TestcaseInfo, seed: Optional[int], test: int
    ) -> None:
        input_path = testcase_info.input_path(seed)
        assert self.inputs[input_path.name][1] == seed
        self.inputs[input_path.name][0].add(test)

    def _add_generate_input_jobs(
        self,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        test: int,
        test_determinism: bool,
    ) -> None:
        input_path = testcase_info.input_path(seed)

        if testcase_info.generation_mode == TestcaseGenerationMode.generated:
            self._add_job(self._generate_input_job(testcase_info, seed), new_last=True)

        if (
            testcase_info.generation_mode == TestcaseGenerationMode.generated
            and test_determinism
        ):
            assert self._env.config.tests.in_gen is not None
            self._add_job(
                generator_test_determinism(
                    self._env, self._env.config.tests.in_gen, testcase_info, seed
                )
            )

        self._check_input_jobs(input_path)

        if self._env.config.tests.validator is not None:
            for t in range(self._env.config.tests_count):
                if not self._env.config.test_sections[t].in_test(input_path.name):
                    continue
                if not self._env.config.test_sections[t].checks_validate:
                    continue

                self._add_job(
                    self._validate(
                        input_path,
                        t,
                    )
                )

    def _validate(self, input_path: IInputPath, test_num: int) -> ValidatorJob:
        assert self._env.config.tests.validator is not None
        assert self._env.config.tests.validator_type is not None

        return VALIDATORS[self._env.config.tests.validator_type](
            self._env, self._env.config.tests.validator, input_path, test_num
        )

    def _generate_input_job(
        self, testcase_info: TestcaseInfo, seed: Optional[int]
    ) -> GenerateInput:
        assert self._env.config.tests.in_gen is not None
        self._gen_inputs_job[seed] = gen_inp = generate_input(
            self._env, testcase_info, seed
        )
        return gen_inp

    def _add_solution_jobs(
        self,
        testcase_info: TestcaseInfo,
        seed: Optional[int],
        test: int,
    ) -> None:
        pass

    def _add_respects_seed_jobs(
        self, testcase_info: TestcaseInfo, seeds: list[int], test: int
    ) -> None:
        assert (
            testcase_info.generation_mode == TestcaseGenerationMode.generated
            and testcase_info.seeded
        )

        if len(seeds) == 1:
            seeds.append(seed := self._get_seed(1, testcase_info))
            self._add_job(self._generate_input_job(testcase_info, seed))

        self._add_job(
            check_seeded := GeneratorRespectsSeed(self._env, testcase_info, *seeds[:2])
        )
        for i in range(2):
            check_seeded.add_prerequisite(self._gen_inputs_job[seeds[i]])

    def _check_input_jobs(self, input_path: IInputPath) -> None:
        self._add_job(
            sanitize_job(self._env, input_path, True),
            new_last=True,
        )

        if self._env.config.limits.input_max_size != 0:
            self._add_job(InputSmall(self._env, input_path))

    def _add_check_output_jobs(
        self,
        output_path: IOutputPath,
    ) -> None:
        self._add_job(
            sanitize_job(self._env, output_path, False),
            prerequisite_name="create-source",
            new_last=True,
        )

        if self._env.config.limits.output_max_size != 0:
            self._add_job(OutputSmall(self._env, output_path))

    def _compute_result(self) -> dict[str, Any]:
        return {
            "input_dataset": list(sorted(self.input_dataset, key=lambda i: i.name)),
            "inputs": {i: (list(sorted(t)), s) for i, (t, s) in self.inputs.items()},
        }


class RunGenerator(TaskJobManager, TestcaseInfoMixin):
    def __init__(self) -> None:
        super().__init__("Run generator")

    def _get_jobs(self) -> list[Job]:
        for sub_num, inputs in self._all_testcases().items():
            for inp in inputs:
                self._add_testcase_info_jobs(inp, sub_num)

        return self._jobs



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2025        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

# !!! BEWARE: This library is backwards incompatible and it's interface can change. !!!

import os
import shutil
import tempfile

from pisek.user_errors import (
    InvalidArgument,
    InvalidOperation,
    TestingFailed,
    NotSupported,
)
from pisek.utils.paths import (
    BUILD_DIR,
    TESTS_DIR,
    INPUTS_LIST,
    INPUTS_SUBDIR,
    OpendataRawPath,
    OpendataInputPath,
    OpendataOutputPath,
)
from pisek.utils.pipeline_tools import run_pipeline
from pisek.config.config_types import GenType
from pisek.config.config_hierarchy import DEFAULT_CONFIG_FILENAME
from pisek.config.task_config import load_config
from pisek.config.config_tools import export_config
from pisek.env.env import TestingTarget
from pisek.jobs.task_pipeline import TaskPipeline
from pisek.task_jobs.data.testcase_info import TestcaseInfo
from pisek.task_jobs.data.data_manager import TEST_SEED

from pisek.opendata.types import OpendataTestcaseInfo, OpendataVerdict
from pisek.opendata.managers import OpendataPipeline


ENV_ARGS = {
    "no_colors": True,
    "no_jumps": True,
}


class Task:
    def __init__(
        self,
        path: str,
        pisek_dir: str | None = None,
        config_filename: str = DEFAULT_CONFIG_FILENAME,
    ) -> None:
        self._path = path
        self._pisek_dir = pisek_dir or os.environ.get("PISEK_DIRECTORY")
        self._config_filename = config_filename

    def test(self, strict: bool = True, disable_cache: bool = False) -> bool:
        try:
            run_pipeline(
                self._path,
                TaskPipeline,
                pisek_dir=self._pisek_dir,
                config_filename=self._config_filename,
                disable_cache=disable_cache,
                strict=strict,
                **ENV_ARGS,
            )
        except TestingFailed:
            return False
        return True

    def build(self, path: str) -> "BuiltTask":
        config = load_config(
            self._path, self._pisek_dir, self._config_filename, False, True
        )
        if config.tests.gen_type == GenType.cms_old:
            raise NotSupported(
                "For opendata tasks, the cms-old generator is not supported."
            )

        if os.path.exists(path):
            if not os.path.isdir(path):
                raise InvalidArgument(f"{path} should be a directory")
            if os.listdir(path):
                raise InvalidOperation(f"{path} should be an empty directory")

        os.makedirs(path, exist_ok=True)
        run_pipeline(
            self._path,
            TaskPipeline,
            pisek_dir=self._pisek_dir,
            config_filename=self._config_filename,
            target=TestingTarget.build,
            disable_cache=True,
            **ENV_ARGS,
        )

        COPIED_PATHS = [
            BUILD_DIR,
            os.path.join(TESTS_DIR, INPUTS_SUBDIR),
            os.path.join(TESTS_DIR, INPUTS_LIST),
        ]

        for task_path in COPIED_PATHS:
            original_path = os.path.join(self._path, task_path)
            full_path = os.path.join(path, task_path)
            if os.path.isdir(original_path):
                shutil.copytree(original_path, full_path)
            elif os.path.isfile(original_path):
                shutil.copy(original_path, full_path)
            else:
                assert False

        export_config(
            self._path,
            self._pisek_dir,
            self._config_filename,
            os.path.join(path, DEFAULT_CONFIG_FILENAME),
        )

        return BuiltTask(path)


class BuiltTask:
    def __init__(self, path: str) -> None:
        self._path = path
        self._config = load_config(path, None, DEFAULT_CONFIG_FILENAME, True, True)

    def _inputs_list(self) -> dict[str, TestcaseInfo]:
        infos = {}
        with open(os.path.join(self._path, TESTS_DIR, INPUTS_LIST)) as f:
            for line in f:
                ti = TestcaseInfo.from_str(line)
                infos[ti.name] = ti
        return infos

    def inputs_list(self) -> dict[int, list[OpendataTestcaseInfo]]:
        result: dict[int, list[OpendataTestcaseInfo]] = {
            test_num: [] for test_num in range(self._config.tests_count)
        }
        for ti in self._inputs_list().values():
            for test_num in result.keys():
                if self._config.test_sections[test_num].in_test(
                    ti.input_path(TEST_SEED).name
                ):
                    result[test_num].append(OpendataTestcaseInfo.from_testcase_info(ti))

        return result

    def get_testcase(
        self,
        name: str,
        test: int,
        seed: int | None,
        input_path: str,
        output_path: str,
    ) -> "Testcase":
        if not 0 <= test <= self._config.tests_count:
            raise InvalidArgument(f"No test with number {test}")

        testcase_infos = self._inputs_list()
        if name not in testcase_infos:
            raise InvalidArgument(f"There is no testcase named '{name}'")
        testcase_info = testcase_infos[name]

        if not self._config.test_sections[test].in_test(
            testcase_info.input_path(TEST_SEED).name
        ):
            raise InvalidArgument(f"Testcase '{name}' is not in test {test}")

        if seed == None and testcase_info.seeded:
            raise InvalidArgument(f"No seed for seeded testcase")
        elif seed is not None and not testcase_info.seeded:
            raise InvalidArgument(f"Seed for unseeded testcase")

        if input_path == output_path:
            raise InvalidArgument("Input and output path must be different")

        return Testcase(self, testcase_info, test, seed, input_path, output_path)


class Testcase:
    def __init__(
        self,
        built_task: BuiltTask,
        info: TestcaseInfo,
        test: int,
        seed: int | None,
        input_path: str,
        output_path: str,
    ) -> None:
        """Use BuiltTask.get_testcase instead."""
        self._built_task = built_task
        self._info = info
        self._test = test
        self._seed = seed
        self._input_dst = input_path
        self._output_dst = output_path

        self.tmp_dir: str | None = None

    @staticmethod
    def _path_present(path: str | None) -> bool:
        return path is not None and os.path.exists(path)

    def _input_path(self) -> OpendataInputPath:
        assert self.tmp_dir is not None
        return OpendataInputPath(self.tmp_dir, self._input_dst)

    def _output_path(self) -> OpendataOutputPath:
        assert self.tmp_dir is not None
        return OpendataOutputPath(self.tmp_dir, self._output_dst)

    def _make_tmp_dir(self) -> None:
        if self.tmp_dir is None:
            self.tmp_dir = tempfile.mkdtemp(prefix="pisek_opendata_")

    def _clear_tmp_dir(self) -> None:
        assert self.tmp_dir is not None
        shutil.rmtree(self.tmp_dir)
        self.tmp_dir = None

    def _run_pipeline(
        self,
        input_needed: bool = False,
        output_needed: bool = False,
        check: bool = False,
        contestant_output: str | None = None,
        clear: bool = False,
    ) -> OpendataVerdict | None:
        output_needed &= not self._path_present(self._output_dst)
        input_needed |= output_needed
        input_needed &= not self._path_present(self._input_dst)

        if input_needed or output_needed or check:
            self._make_tmp_dir()
            assert self.tmp_dir is not None

            contestant_output_path: OpendataRawPath | None = None
            if contestant_output is not None:
                contestant_output_path = OpendataRawPath(
                    self.tmp_dir, contestant_output
                )

            pipeline = OpendataPipeline(
                input_needed,
                output_needed,
                check,
                self._input_path(),
                self._info,
                self._test,
                self._seed,
                self._output_path(),
                contestant_output_path,
            )
            run_pipeline(
                self._built_task._path,
                lambda _: pipeline,
                disable_cache=True,
                **ENV_ARGS,
            )
            if clear:
                self._clear_tmp_dir()

        if check:
            return pipeline.verdict
        return None

    def gen_input(self, clear: bool = True) -> None:
        self._run_pipeline(input_needed=True, clear=clear)

    def gen_output(self, clear: bool = True) -> None:
        self._run_pipeline(output_needed=True, clear=clear)

    def check(self, contestant_output_path: str, clear: bool = True) -> OpendataVerdict:
        assert self._built_task._config.tests.judge_needs_in is not None
        assert self._built_task._config.tests.judge_needs_out is not None

        res = self._run_pipeline(
            input_needed=self._built_task._config.tests.judge_needs_in,
            output_needed=self._built_task._config.tests.judge_needs_out,
            check=True,
            contestant_output=contestant_output_path,
            clear=clear,
        )
        assert res is not None
        return res



import unittest

from util import TestFixtureVariant, modify_config


class TestOddStub(TestFixtureVariant):
    @property
    def fixture_path(self) -> str:
        return "fixtures/odd_stub/"


class TestBigInput(TestOddStub):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config.add_section("limits")
            raw_config["limits"]["input_max_size"] = "1"

        modify_config(self.task_dir, modification_fn)


class TestBigOutput(TestOddStub):
    def expecting_success(self) -> bool:
        return False

    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config.add_section("limits")
            raw_config["limits"]["output_max_size"] = "1"

        modify_config(self.task_dir, modification_fn)


if __name__ == "__main__":
    unittest.main(verbosity=2)



from pisek.config.config_types import ValidatorType
from pisek.task_jobs.validator.validator_base import ValidatorJob
from pisek.task_jobs.validator.simple_validator import Simple0Validate, Simple42Validate

VALIDATORS: dict[ValidatorType, type[ValidatorJob]] = {
    ValidatorType.simple_0: Simple0Validate,
    ValidatorType.simple_42: Simple42Validate,
}



import os
from typing import Optional

from pisek.config.update_config import update_config
from pisek.config.config_hierarchy import (
    ConfigHierarchy,
    GLOBAL_DEFAULTS,
    new_config_parser,
)
from pisek.config.task_config import load_config, TaskConfig


def update_and_replace_config(
    task_path: str,
    pisek_directory: Optional[str],
    config_filename: str,
) -> None:
    load_config(task_path, pisek_directory, config_filename, suppress_warnings=True)

    config_path = os.path.join(task_path, config_filename)
    config = new_config_parser()
    config.read(config_path)
    update_config(config, task_path=task_path, infos=False)
    with open(config_path, "w") as f:
        config.write(f, space_around_delimiters=False)


def export_config(
    task_path: str,
    pisek_directory: str | None,
    config_filename: str,
    exported_config_path: str,
) -> None:
    config_hierarchy = ConfigHierarchy(
        task_path, False, pisek_directory, config_filename
    )
    TaskConfig.load_dict(config_hierarchy)

    config = new_config_parser()
    config.add_section("task")
    config["task"]["version"] = "v3"

    for config_val in config_hierarchy.loaded_values:
        if not config_val.internal and config_val.config != GLOBAL_DEFAULTS:
            if not config.has_section(config_val.section):
                config.add_section(config_val.section)
            if config_val.key is not None:
                config[config_val.section][config_val.key] = config_val.value

    with open(exported_config_path, "w") as f:
        config.write(f, space_around_delimiters=False)



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from typing import Any, TYPE_CHECKING

from pisek.env.context import ContextModel


class BaseEnv(ContextModel):
    """
    Collection of environment variables which logs whether each variable was accessed.
    """

    def __init__(self, *args, **kwargs) -> None:
        super().__init__(*args, **kwargs)
        self._accessed: set[str] = set()
        self._logging: bool = True

        self._direct_subenvs: list[str] = []
        self._dict_subenvs: list[str] = []
        self._precompute_subenvs()

    if not TYPE_CHECKING:
        # XXX: This is a bit black-magicky because of efficiency
        # Be careful when touching this
        def __getattribute__(self, item: str) -> Any:
            if not item.startswith("_") and hasattr(self, "_accessed"):
                super().__getattribute__("_accessed").add(item)
            return ContextModel.__getattribute__(self, item)

    def fork(self):
        """Make copy of this env with no accesses logged."""
        model = self.model_copy(deep=True)
        model.clear_accesses()
        return model

    def _precompute_subenvs(self):
        def is_env(item: Any) -> bool:
            return isinstance(item, BaseEnv)

        for key in type(self).model_fields:
            item = getattr(self, key)
            if is_env(item):
                self._direct_subenvs.append(key)
            elif isinstance(item, dict) and any(map(is_env, item.values())):
                assert all(map(is_env, item.values()))
                self._dict_subenvs.append(key)

    def clear_accesses(self) -> None:
        """Remove all logged accesses."""
        accessed = super().__getattribute__("_accessed")
        for direct_subenv in super().__getattribute__("_direct_subenvs"):
            if direct_subenv in accessed:
                super().__getattribute__(direct_subenv).clear_accesses()

        for dict_subenv in super().__getattribute__("_dict_subenvs"):
            if dict_subenv in accessed:
                subenv = super().__getattribute__(dict_subenv)
                for subitem in subenv.values():
                    subitem.clear_accesses()

        super().__getattribute__("_accessed").clear()

    def get_accessed(self) -> set[tuple[str, ...]]:
        """Get all accessed field names in this env (and all subenvs)."""
        cls = type(self)
        accessed = set()
        self._accessed &= set(cls.model_fields) | set(cls.model_computed_fields)
        for key in self._accessed:
            item = getattr(self, key)
            if isinstance(item, BaseEnv):
                accessed |= {(key, *subkey) for subkey in item.get_accessed()}
            elif isinstance(item, dict) and all(
                isinstance(val, BaseEnv) for val in item.values()
            ):
                accessed |= {
                    (key, dict_key, *subkey)
                    for dict_key, subenv in item.items()
                    for subkey in subenv.get_accessed()
                }
            else:
                accessed.add((key,))

        return accessed

    def get_compound(self, key: tuple[str, ...]) -> Any:
        """Get attribute that may be nested deeper and indexed."""
        obj = self
        for key_part in key:
            if isinstance(obj, BaseEnv):
                obj = getattr(obj, key_part)
            elif isinstance(obj, dict):
                obj = obj[type(list(obj.keys())[0])(key_part)]
            else:
                raise ValueError(f"Can't get compound key on type '{type(obj)}'.")
        return obj



#!/usr/bin/env python3
# Validator based on Kasiopea's validator template from Pali Madaj.
import sys
import argparse


def main(diff):
    BOUNDS = [(-1e18, 1e18), (0, 1e9), (-1e9, 1e9), (-1e18, 1e18)]
    assert 0 <= diff < len(BOUNDS)
    read_values(2, *BOUNDS[diff])


# ----------------- Cast nize jsou pomocne funkce, snad nebude treba upravovat -----------------

line_number = 0


def read_line():
    global line_number
    line_number += 1
    return input()


def fail(message):
    global line_number
    print(message + " (on line {})".format(line_number), file=sys.stderr)
    sys.exit(1)


def read_values(count=None, minimum=None, maximum=None, value_type=int):
    try:
        line = read_line()
    except EOFError:
        fail("Unexpected end of file.")

    try:
        numbers = list(map(value_type, line.split(" ")))
    except ValueError as err:
        fail(str(err))

    if count is not None and len(numbers) != count:
        fail(
            "The number of values was {} and should have been {}.".format(
                len(numbers), count
            )
        )
    if minimum is not None and any(x < minimum for x in numbers):
        fail("A value was smaller than {}.".format(minimum))
    if maximum is not None and any(x > maximum for x in numbers):
        fail("A value was greater than {}.".format(maximum))

    return numbers


def expect_eof():
    koniec = False
    try:
        input()
    except EOFError:
        koniec = True
    if not koniec:
        fail("The file continues, but it should have ended.")


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Loads an input from stdin and validates it."
    )
    parser.add_argument("test", type=int, help="test number (0-3)")
    args = parser.parse_args()
    main(args.test)
    expect_eof()
    exit(42)



"""
A module for testing pisek itself. The strategy is to take a functioning
fixture of a task and then break it in various small ways to see whether
pisek catches the problem.
"""

from decimal import Decimal
import io
import tempfile
import os
import shutil
import unittest

from util import TestFixture, modify_config

from pisek.user_errors import UserError
from pisek.task_jobs.data.testcase_info import TestcaseGenerationMode
from pisek.task_jobs.solution.solution_result import Verdict
from pisek.opendata.types import OpendataTestcaseInfo, OpendataVerdict
from pisek.opendata.lib import Task, BuiltTask


class TestFixtureOpendata(TestFixture):
    def expecting_success(self) -> bool:
        return True

    def modify_task(self) -> None:
        pass

    @property
    def _testcase_files(self) -> list[str]:
        files = []
        for file in [self.input_path, self.output_path, self.contestant_path]:
            if file is not None:
                files.append(file)
        return files

    def created_files(self):
        return [os.path.basename(file) for file in self._testcase_files]

    def runTest(self) -> None:
        if not self.fixture_path:
            return

        self.log_files()
        task = Task(self.task_dir)

        self.built_task: BuiltTask

        self.input_path: str | None = None
        self.output_path: str | None = None
        self.contestant_path: str | None = None

        @unittest.mock.patch("sys.stdout", new_callable=io.StringIO)
        @unittest.mock.patch("sys.stderr", new_callable=io.StringIO)
        def run(*args) -> bool:
            try:
                self.modify_task()
                self._build_task_dir = tempfile.mkdtemp(prefix="pisek-test_")
                self.built_task = task.build(self._build_task_dir)
                self.run_opendata_test()
                return True
            except UserError as e:
                print(e)
                return False

        self.assertEqual(run(), self.expecting_success())

        for file in self._testcase_files:
            self.assertTrue(os.path.isfile(file))

        self.check_end_state()
        self.check_files()

    def tearDown(self) -> None:
        if not self.fixture_path:
            return

        assert self._build_task_dir.startswith(
            "/tmp"
        ) or self._build_task_dir.startswith("/var")
        shutil.rmtree(self._build_task_dir)

    def check_end_state(self):
        # Here we can verify whether some conditions hold when Pisek finishes,
        # making sure that the end state is reasonable
        pass

    def run_opendata_test(self):
        pass


class TestSumKasiopeaOpendataBuild(TestFixtureOpendata):
    @property
    def fixture_path(self) -> str:
        return "fixtures/sum_kasiopea/"

    def init_testcase(self, name: str) -> None:
        self.input_path = os.path.join(self.task_dir, f"{name}.opendata.in")
        self.output_path = os.path.join(self.task_dir, f"{name}.opendata.out")
        self.testcase = self.built_task.get_testcase(
            name, int(name), int("deadbeef", 16), self.input_path, self.output_path
        )

    def create_contestant_file(self, content: str | bytes) -> None:
        self.contestant_path = os.path.join(self.task_dir, "02.out")

        if isinstance(content, bytes):
            with open(self.contestant_path, "xb") as f:
                f.write(content)
        else:
            with open(self.contestant_path, "x") as f:
                f.write(content)


class TestSumKasiopeaOpendataListInputs(TestSumKasiopeaOpendataBuild):
    def run_opendata_test(self):
        self.assertEqual(
            self.built_task.inputs_list(),
            {
                0: [
                    OpendataTestcaseInfo(
                        TestcaseGenerationMode.static, "sample", 1, False
                    )
                ],
                1: [
                    OpendataTestcaseInfo(
                        TestcaseGenerationMode.generated, "01", 1, True
                    )
                ],
                2: [
                    OpendataTestcaseInfo(
                        TestcaseGenerationMode.generated, "02", 1, True
                    )
                ],
            },
        )


class TestSumKasiopeaOpendataSequential(TestSumKasiopeaOpendataBuild):
    def run_opendata_test(self) -> None:
        self.init_testcase("01")
        assert self.input_path is not None
        assert self.output_path is not None

        self.testcase.gen_input()
        self.assertTrue(os.path.exists(self.input_path))
        self.testcase.gen_output()
        self.assertTrue(os.path.exists(self.output_path))
        self.assertEqual(
            self.testcase.check(self.output_path),
            OpendataVerdict(Verdict.ok, None, Decimal(4), None, None),
        )


class TestSumKasiopeaOpendataCheckRightaway(TestSumKasiopeaOpendataBuild):
    def run_opendata_test(self) -> None:
        self.init_testcase("02")
        self.assertEqual(
            self.testcase.check(os.path.join(self.task_dir, "sample.out")),
            OpendataVerdict(Verdict.wrong_answer, None, Decimal(0), None, None),
        )


class TestSumKasiopeaOpendataCheckBinary(TestSumKasiopeaOpendataBuild):
    def run_opendata_test(self):
        self.init_testcase("02")
        self.create_contestant_file(b"\x07\n")

        self.assertEqual(
            self.testcase.check(os.path.join(self.task_dir, self.contestant_path)),
            OpendataVerdict(
                Verdict.normalization_fail,
                "File contains non-printable character (code 7 at position 0)",
                Decimal(0),
                None,
                None,
            ),
        )


class TestSumKasiopeaOpendataJudge(TestSumKasiopeaOpendataBuild):
    def modify_task(self) -> None:
        def modification_fn(raw_config):
            raw_config["tests"]["out_check"] = "judge"
            raw_config["tests"]["out_judge"] = "judge"
            raw_config["tests"]["judge_type"] = "opendata-v2"

        modify_config(self.task_dir, modification_fn)

    def run_opendata_test(self):
        self.init_testcase("02")
        self.create_contestant_file("0\n")

        self.assertEqual(
            self.testcase.check(os.path.join(self.task_dir, self.contestant_path)),
            OpendataVerdict(
                Verdict.wrong_answer,
                "Wrong answer",
                Decimal(0),
                None,
                None,
            ),
        )


if __name__ == "__main__":
    unittest.main()



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

from random import Random
import string

from pisek.env.env import Env
from pisek.utils.paths import TaskPath
from pisek.task_jobs.task_job import TaskJob


def randword(length: int, rand_gen: Random):
    letters = string.ascii_lowercase
    return "".join(rand_gen.choice(letters) for _ in range(length))


class Invalidate(TaskJob):
    """Abstract Job for invalidating an output."""

    def __init__(
        self, env: Env, name: str, from_file: TaskPath, to_file: TaskPath, seed: int
    ) -> None:
        super().__init__(env, name)
        self.seed = seed
        self.from_file = from_file
        self.to_file = to_file

    def _select_line_index(self, rand_gen: Random, lines_len: int) -> int:
        assert lines_len > 0
        if lines_len <= 2 or rand_gen.randint(1, 10) == 1:
            return rand_gen.randint(0, min(1, lines_len - 1))
        else:
            return rand_gen.randint(2, lines_len - 1)


class Incomplete(Invalidate):
    """Makes an incomplete output."""

    def __init__(
        self, env: Env, from_file: TaskPath, to_file: TaskPath, seed: int
    ) -> None:
        super().__init__(
            env,
            f"Incomplete {from_file:n} -> {to_file:n} (seed {seed:x})",
            from_file,
            to_file,
            seed,
        )

    def _run(self):
        with self._open_file(self.from_file) as f:
            lines = f.readlines()

        rand_gen = Random(self.seed)
        if lines:
            lines = lines[: rand_gen.randint(0, len(lines) - 1)]

        with self._open_file(self.to_file, "w") as f:
            f.write("".join(lines))


class BlankLine(Invalidate):
    """Makes a line in the output blank."""

    def __init__(
        self, env: Env, from_file: TaskPath, to_file: TaskPath, seed: int
    ) -> None:
        super().__init__(
            env,
            f"Blank line {from_file:n} -> {to_file:n} (seed {seed:x})",
            from_file,
            to_file,
            seed,
        )

    def _run(self):
        with self._open_file(self.from_file) as f:
            lines = f.readlines()

        rand_gen = Random(self.seed)
        if lines:
            lines[self._select_line_index(rand_gen, len(lines))] = "\n"

        with self._open_file(self.to_file, "w") as f:
            f.write("".join(lines))


class ChaosMonkey(Invalidate):
    """Tries to break judge by generating nasty output."""

    def __init__(self, env, from_file: TaskPath, to_file: TaskPath, seed: int) -> None:
        super().__init__(
            env,
            f"ChaosMonkey {from_file:n} -> {to_file:n} (seed {seed:x})",
            from_file,
            to_file,
            seed,
        )

    def _run(self):
        rand_gen = Random(self.seed)

        NUMBER_MODIFIERS = [
            lambda _: 0,
            lambda x: int(x) + 1,
            lambda x: int(x) - 1,
            lambda x: -int(x),
            lambda x: int(x) + rand_gen.randint(1, 9) / 10,
        ]
        CREATE_MODIFIERS = [
            lambda _: rand_gen.randint(0, int(1e5)),
            lambda _: rand_gen.randint(-int(1e5), -1),
            lambda _: rand_gen.randint(0, int(1e18)),
            lambda _: rand_gen.randint(-int(1e18), -1),
            lambda _: randword(rand_gen.randint(1, 10), rand_gen),
        ]
        CHANGE_MODIFIERS = [
            lambda x: f"{x} {x}",
            lambda _: "",
            lambda x: randword(len(x), rand_gen),
            lambda x: randword(len(x) + 1, rand_gen),
            lambda x: randword(len(x) - 1, rand_gen),
        ]

        lines = []
        with self._open_file(self.from_file) as f:
            for line in f.readlines():
                lines.append(line.rstrip("\n").split(" "))

        if len(lines) == 0:
            lines = [[str(rand_gen.choice(CREATE_MODIFIERS)(""))]]
        else:
            line = self._select_line_index(rand_gen, len(lines))
            token = rand_gen.randint(0, len(lines[line]) - 1)

            modifiers = CREATE_MODIFIERS + CHANGE_MODIFIERS
            try:
                int(lines[line][token])
                modifiers += NUMBER_MODIFIERS
            except ValueError:
                pass
            lines[line][token] = str(rand_gen.choice(modifiers)(lines[line][token]))

        with self._open_file(self.to_file, "w") as f:
            for line in lines:
                f.write(" ".join(line) + "\n")


class TrailingString(Invalidate):
    def __init__(
        self, env: Env, from_file: TaskPath, to_file: TaskPath, seed: int
    ) -> None:
        super().__init__(
            env,
            f"Trailing string {from_file:n} -> {to_file:n} (seed {seed:x})",
            from_file,
            to_file,
            seed,
        )

    def _run(self):
        with self._open_file(self.from_file) as f:
            lines = f.readlines()

        # TODO: Find some more permanent solution (#545)
        if lines and lines[-1].endswith("\n"):
            lines[-1] += "\n"

        rand_gen = Random(self.seed)
        lines.append(randword(60, rand_gen) + "\n")

        with self._open_file(self.to_file, "w") as f:
            f.write("".join(lines))



# pisek  - Tool for developing tasks for programming competitions.
#
# Copyright (c)   2019 - 2022 Václav Volhejn <vaclav.volhejn@gmail.com>
# Copyright (c)   2019 - 2022 Jiří Beneš <mail@jiribenes.com>
# Copyright (c)   2020 - 2022 Michal Töpfer <michal.topfer@gmail.com>
# Copyright (c)   2022        Jiří Kalvoda <jirikalvoda@kam.mff.cuni.cz>
# Copyright (c)   2023        Daniel Skýpala <daniel@honza.info>

# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# any later version.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.
import os

from pisek.utils.text import pad

MSG_LEN = 25

try:
    terminal_width, terminal_height = os.get_terminal_size()
except OSError:
    terminal_width, terminal_height = 100, 24

TARGET_LINE_WIDTH = min(terminal_width, 120)
LINE_SEPARATOR = "—" * terminal_width + "\n"


def separator_text(text: str):
    dashes = TARGET_LINE_WIDTH - len(text) - 2
    return "-" * (dashes // 2) + " " + text + " " + "-" * ((dashes + 1) // 2)


def right_aligned_text(left: str, right: str, offset: int = 0):
    return pad(left, TARGET_LINE_WIDTH - len(right) + offset - 1) + " " + right
